{-----------------------------------------------------------------------------
 Unit Name: FGammaPicker
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 Purpose: Color Picker
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}


unit FGammaPicker;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, VectorsAndUtil, ComCtrls;
const
  cLastGamma = 1;
  cLastPtSpline1 = 5;
  cLastPtSpline2= 7;
  cLastPtSpline = 12;
  cQ= 4;
  cGammaN = 511; {last color in the gammas; must be an odd number}
  kMedPltIndex =  (cGammaN+1) div 2;
  cWheelRadius = 80;
  cMaxLAmplitude= 80;
  cMaxComp = 255;
type
    TColorArray = array[0..65535] of TColor;
    pTColorArray = ^TColorArray;
    Gamma = array[0..cGammaN] of TColor;
    gammaList= array[0..cLastGamma] of Gamma;
    InterpPoints= array[0..cLastPtSpline] of Vector3;
    Plt = array[0..cGammaN] of TColor;

type
   {The gamma picker consists of two TPaintBoxes, the hue-saturation called
    'HsBox' and the illumination, ligthness or brightness one, called 'LBox'.
    Suppose that we have a curve in the 3-dimensional space whose dimensions
    are Hue, Saturation and Lightness, and denote it by alpha(t), for
    t between 0 and cGammaN; and suppose that we have selected some
    points alpha(t_1),...,alpha(t_m), that we shall call "control points",
    where the values 't_1.,,,.t_m' are equally spaced. The 3 coordinates of
    the point alpha(n), where 'n' is an integer between 0 and cGammaN, define a color. So that we have a
    gamma of cGammaN + 1 colors defined by that curve. That curve projects
    onto the plane of Hue-Saturation, and it is supposed that this projection
    goes inside the circle of hue and saturation of the HSL system. We identify
    with little circles the control points of the curve: they may be moved by
    the user. In the same manner we draw the graph of the third coordinate in
    the LBox. In abscissas we put the value of 't' and in ordinates the value of
    the lightness or third coordinate of 'alpha(t)'. Since for each 't', the
    two first coordinates of alpha(t) determine a hue,H , and a saturation, S,
    we draw along the vertical line of pixels with abscissa 't' the color
    that correspond to hue H, saturation S and lightness given by the ordinate,
    that is by the third component of alpha(t). The control points are also
    highlighted in the curve of LBox, and they may be moved by the user. As
    the user moves the control points, the program updates in real time the
    color of the surface, if it corresponds to the color tag chosen in the
    gammaPicker form.}


  TGammaPicker = class(TForm)
    GammaTabs: TTabControl;
    HSBox: TPaintBox;
    LBox: TPaintBox;
    procedure DispatchBackgroundChange;
    procedure DispatchGammaChange;
    procedure MouseDownInHSBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseDownInLBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseMoveInHSBox(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MouseMoveInLBox(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MouseUpInHSBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseUpInLBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnChangeGammaTabs(Sender: TObject);
    procedure OnPaintPicker(Sender: TObject);
    procedure OnShowGPicker(Sender: TObject);
    procedure PaintHSBox(Sender: TObject);
    procedure PaintLBox(Sender: TObject);
  private
    own: TComponent;
    HSBm, LBm: TBitmap;
    halfLHeight: Integer;
    lastHSPt: TPoint;
    trackingHS, trackingL, translating, spinning: Boolean;
    trackingPts: array[0..cLastPtSpline] of Boolean;
    bb: array[-cQ..cQ] of Double;
    points: array[0..1,0..cLastPtSpline+ 2*cQ + 4] of Vector3;
    procedure BuildBackgrPlt(var pal: Plt; h1, h2: Double);
    procedure CompleteGridAndAdjustItHSandL;
    procedure CompleteGridAndAdjustItL;
    procedure ComputeAndDrawHSLGamma(p: Vector2; bm: TBitmap; column: Integer);
    procedure DrawWheel;
    function EvaluateBSpline(t: Double): Vector3;
    function FromHSLToRGB(hs: Vector2; b: Double): TColor;
    function FromWheelToHS(x,y: Double): Vector2;
    function PtInWheel(var r: Double; h,v: Integer): Boolean;
  public
    tg: Integer;
    LRect, HSRect: TRect;
    Lctr, HSctr: Vector2Int;
    iPts: array[0..cLastGamma] of InterpPoints;
    phi: array [0..cLastGamma] of Double;
    procedure ComputeGamma(gtag: Integer);
    constructor Create(aOwner: TComponent); override;
    procedure DrawBSplineCurveHSandL;
    procedure DrawBSplineCurveL;
    procedure Initiate(gtag: Integer; notify: Boolean);
    destructor Destroy; override;
  end;

  {For the use of B-splines in interpolation, consult the notes
  "AproxFuncBSplinesEnglish.pdf" included in this folder}


  var
    gPicker: TGammaPicker;
    gammas: GammaList;
    bckgPal: Plt;

implementation
{$R *.dfm}
  uses math,
    FImmersionR3R6, ResourceStrings, FDrawParamSurf;

  const
    cTwoPi: double = 6.2831853071795864;

procedure TGammaPicker.BuildBackgrPlt(var pal: Plt; h1, h2: Double);
{build a RGB color palette 'pal' that has its brightest value at the center,
and goes downward towards black with hue and saturation 'h2' and upward
towards black with hue and saturation 'h1'}

const
  cMargin= 0.05;
var
  hs1, hs2: Vector2;
  step: Double;
  i: Integer;
begin
  hs1[1]:= h1; {hue of first color}
  hs1[2]:= 1; {saturation of first color}
  hs2[1]:= h2; {hue of last color}
  hs2[2]:= 0.9; {saturation of last color}
  step:= (1-2*cMargin)/(kMedPltIndex-1);
  for i:= 0 to kMedPltIndex-1 do begin
    {with this choice of 'i' and 'step', all the entries of the palette are
     assigned, the least light is 'cMargin' and the greatest '1-cMargin'}
    pal[i]:= FromHSLToRGB(hs1,cMargin+i*step);
    pal[cGammaN-i]:= FromHSLToRGB(hs2,cMargin+i*step);
  end;
end;



procedure TGammaPicker.CompleteGridAndAdjustItHSandL;
{Prepare internal control points of the B-spline curve in the HSBox and in the
 L Box. See the memory "AproxFuncBSplinesEnglish.pdf" included in this folder.
 The values that are interpolated are the coordinates}
  var
    i, j, h: Integer;
begin
  for i:= 0 to cLastPtSpline do
    for j:= 1 to 3 do
      points[1,cQ+1+i,j]:= ipts[tg,i,j];
  for i:= 0 to cQ do
    for j:= 1 to 3 do
      points[1,i,j]:= 2*points[1,cQ+1,j] - points[1,2*(cQ+1)-i,j];
  h:= cQ + cLastPtSpline + 1;
  for i:= h+1 to h + cQ + 2  do
    for j:= 1 to 3 do
        points[1,i,j]:= 2*points[1,h,j] - points[1,2*h-i,j];
  for i:= cQ-2 to cQ + cLastPtSpline + 1 do
    for j:= 1 to 3 do begin
      points[0,i,j]:= 0;
      for h:= -cQ to cQ do
        points[0,i,j]:= points[0,i,j] + bb[h]*points[1,i+2+h,j];
    end;
end;

procedure TGammaPicker.CompleteGridAndAdjustItL;
{Prepare internal control points of the B-spline curve in the LBox. See the
 memory "AproxFuncBSplinesEnglish.pdf" included in this folder}
  var
    i, h: Integer;
begin
  for i:= 0 to cLastPtSpline do
    points[1,cQ+1+i,3]:= ipts[tg,i,3];
  for i:= 0 to cQ do
    points[1,i,3]:= 2*points[1,cQ+1,3] - points[1,2*(cQ+1)-i,3];
  h:= cQ + cLastPtSpline + 1;
  for i:= h+1 to h + cQ + 2  do
    points[1,i,3]:= 2*points[1,h,3] - points[1,2*h-i,3];
  for i:= cQ-2 to cQ + cLastPtSpline + 1 do begin
    points[0,i,3]:= 0;
    for h:= -cQ to cQ do
      points[0,i,3]:= points[0,i,3] + bb[h]*points[1,i+2+h,3];
    end;
end;

procedure TGammaPicker.ComputeAndDrawHSLGamma(p: Vector2;
               bm: TBitmap; column: Integer);
{Suppose that p[1], p[2] are the coordinates of the projection of the point
 alpha(t_0) into the HSBox and that 't_0' corresponds to the column of pixels
 'pix' such that 'pix.x = column'. Then , all the column is drawn in the bitmap
 'bm' that usually will be the bitmap of the LBox}
var
  b, step: Double;
  line: pTColorArray;
  j: Integer;
begin
  step:= 1/(LRect.Bottom-1);
  for j:= 0 to LRect.Bottom-1 do begin
    b:= 1 - j*step; {relative value of lightness at ordinate 'j'}
    line:= LBm.ScanLine[j];
    line[column]:= FromHSLToRGB(p,b); {put calculated RGB color to pixel [column,j] }
  end;
end;

procedure TGammaPicker.ComputeGamma(gtag: Integer);
  var
    i, lg: Integer;
    step, t: Double;
    p: Vector3;
    hs: Vector2;
    cl: TColor;
begin
  tg:= gtag;
  {Prepare the abstract control points of the B-Spline}
  CompleteGridAndAdjustItHSandL;

  lg:= High(gammas[tg]);
  step:= 1/lg;
  for i:= 1 to lg do begin
    t:= i*step;
    p:= EvaluateBSpline(t); {so we have the coordinates of the color in
                             the HSBox and LBox}
    hs:= FromWheelToHS(p[1], p[2]); {Obtain hue and saturation from (p[1],p[2])}

    cl:= FromHSLToRGB(hs,(LCtr[2] - p[3])/(LRect.bottom-1));
       {hue and saturation, together with the lightness that corresponds to 'p[3]',
        are passed to 'FromHSLToRGB' for conversion to a RGB color}
    gammas[tg,i]:= cl;
  end
end;

constructor TGammaPicker.Create(aOwner: TComponent);
  var
    a: Double;
    i: Integer;
begin
  inherited create(aOwner);
  own:= owner;
  caption:= ePalettes;

  {initialize canvas of the TPaintBox named LBox (the L is for
   lightness), and the bitmap
   LBm that we use to drawing its content before showing it in
   the window}
  LBox.canvas.Brush.Color:= clWhite;
  LBox.canvas.Pen.Color:= clWhite;
  LBox.Canvas.Brush.Style:= bsSolid;
  LBm:= TBitmap.Create;
  LBm.Width:= LBox.Width;
  LBm.Height:= LBox.Height;
  LBm.PixelFormat:= pf32bit;
  LBm.Canvas.Brush.Style:= bsSolid;
  LBm.Canvas.Brush.Color:= clWhite;
  SetRect(LRect,0,0,LBox.Width,LBox.Height);
  halfLHeight := LBox.Height div 2 -1;

  {Build the tabs in the color gamma picker}
  GammaTabs.Tabs.Clear;
  GammaTabs.Tabs.Add(eStdObverse);
  GammaTabs.Tabs.Add(eStdReverse);

  {initialize canvas of the TPaintBox named HSBox (the HS is
   for hue and saturation), and the bitmap HSBm that we use to
   drawing its content before showing it in the window}
  HSBox.canvas.Brush.Color:= clWhite;
  HSBox.canvas.Pen.Color:= clWhite;
  HSBox.Canvas.Brush.Style:= bsSolid;
  HSBm:= TBitmap.Create;
  HSBm.Width:= HSBox.Width;
  HSBm.Height:= HSBox.Height;
  HSBm.PixelFormat:= pf32bit;
  HSBm.Canvas.Brush.Style:= bsSolid;
  HSBm.Canvas.Brush.Color:= clWhite;
  SetRect(HSRect,0,0,HSBox.Width, HSBox.Height);

  {the centers of HSBox and LBox}
  HSctr[1]:= HSBox.Width div 2;
  HSctr[2]:= HSBox.Height div 2;
  Lctr[1]:= LBox.Width div 2;
  Lctr[2]:= LBox.Height div 2;

  {initialize constants used in the construction of the control points of
  the B-spline curve}
  a := - 1/(2+sqrt(3));
  bb[0]:= 3/(2+a);
  for i:= 1 to cQ do begin
    bb[i]:= bb[i-1]*a;
    bb[-i]:= bb[i];
  end;

  for i:= 0 to cLastGamma do
    phi[i]:= 0;
  DrawWheel;
  translating:= False;
  spinning:= False;
  BuildBackgrPlt(bckgPal, 1/6, 2/3);
end;

destructor TGammaPicker.Destroy;
begin
  LBm.Free;
  HSBm.Free;
  inherited destroy;
end;

procedure TGammaPicker.DispatchBackgroundChange;
{Apply the change of the background color of tab 'tg'}
begin
  gammas[1-tg,0]:= gammas[tg,0];
  ApplyGammaToMap;
  ImmersionR3R6.MainBox.canvas.Draw(0,0,bmp);
end;

procedure TGammaPicker.DispatchGammaChange;
{Apply the change of the gamma of tab 'tg'}
begin
  if ImmersionR3R6.alreadyDrawn then begin
    ApplyGammaToMap;
    ImmersionR3R6.MainBox.canvas.Draw(0,0,bmp);
  end;
end;

procedure TGammaPicker.DrawBSplineCurveHSandL;
{Draw the B-splines curves that define the gamma in the HSBox and in the LBox,
 and change color of the LBox accordingly}
  var
    i: Integer;
    step, t: Double;
    p: Vector3;
    hs: Vector2;
begin
  HSBox.Canvas.Pen.Color:= clWhite;
  HSBox.Canvas.Pen.Mode:= pmCopy;
  CompleteGridAndAdjustItHSandL; {prepare abstract control points of B-splines}
  step:= 1/LRect.right;
  t:= step;
  p:= EvaluateBSpline(t);
  hs:= FromWheelToHS(p[1], p[2]);
  ComputeAndDrawHSLGamma(hs,  LBm, 1); {draw the column 1 in the LBox}
  HSBox.Canvas.moveto(round(p[1]),round(p[2])); {move to the first curve point}
  for i:= 2 to LRect.right-1 do begin {for each pixel abscissa in the LBox, compute
       the corresponding 't' value, the value of the three components of the B-spline
       curve, 'p', draw line in HSBox to the pixel [round(p[1]),round(p[2])],etc.}
    t:= i*step;
    p:= EvaluateBSpline(t);
    if odd(i) then {do not make segments too little}
      HSBox.Canvas.lineto(round(p[1]),round(p[2]));
    hs:= FromWheelToHS(p[1], p[2]);
    ComputeAndDrawHSLGamma(hs, LBm, i); {draw column in the LBox bitmap.}
  end;
  LBox.Canvas.Draw(0,0,LBm);  {draw LBox}
  {complete the curve in HSBox to its last point}
  HSBox.canvas.lineto(round(ipts[tg,cLastPtSpline,1]),round(ipts[tg,cLastPtSpline,2]));

  HSBox.Canvas.Pen.Mode:= pmCopy;
  {Draw a little square around last point in the curve in HSBox}
  HSBox.Canvas.FillRect(Rect(round(ipts[tg,cLastPtSpline,1]-3),round(ipts[tg,cLastPtSpline,2]-3),
       round(ipts[tg,cLastPtSpline,1]+3),round(ipts[tg,cLastPtSpline,2]+3)));

  {Draw a circle around first point in the curve in HSBox}
  HSBox.Canvas.Ellipse(round(ipts[tg,0,1]-4),round(ipts[tg,0,2]-4),round(ipts[tg,0,1]+4),round(ipts[tg,0,2]+4));

  {Frame the little square around last point in the curve in HSBox}
  HSBox.canvas.Brush.Color:= clBlack;
  HSBox.Canvas.moveto(round(ipts[tg,cLastPtSpline,1]-3),round(ipts[tg,cLastPtSpline,2]-3));
  HSBox.Canvas.lineto(round(ipts[tg,cLastPtSpline,1]-3),round(ipts[tg,cLastPtSpline,2]+3));
  HSBox.Canvas.lineto(round(ipts[tg,cLastPtSpline,1]+3),round(ipts[tg,cLastPtSpline,2]+3));
  HSBox.Canvas.lineto(round(ipts[tg,cLastPtSpline,1]+3),round(ipts[tg,cLastPtSpline,2]-3));
  HSBox.Canvas.lineto(round(ipts[tg,cLastPtSpline,1]-3),round(ipts[tg,cLastPtSpline,2]-3));

  {Draw a little circle around each of the other control points in HSBox}
  for i:= 1 to cLastPtSpline-1 do
    HSBox.Canvas.Ellipse(round(ipts[tg,i,1]-3),round(ipts[tg,i,2]-3),round(ipts[tg,i,1]+3),round(ipts[tg,i,2]+3));
  DrawBSplineCurveL;
end;

procedure TGammaPicker.DrawBSplineCurveL;
{Draw the B-splines curve that define the gamma in the LBox}
  var
    i, absc: Integer;
    step, t: Double;
    p: Vector3;
begin
  CompleteGridAndAdjustItL;
  step:= 2/LRect.right;
  t:= 0;
  p:= EvaluateBSpline(t);
  LBox.Canvas.Pen.Color:= clWhite;
  LBox.Canvas.Pen.Mode:= pmCopy;

  {draw the graph of the third component of alpha in LBox}
  LBox.Canvas.moveto(0,round(LCtr[2]+ p[3]));
  for i:= 1 to LRect.right div 2-1 do begin
    t:= i*step;
    p:= EvaluateBSpline(t);
    LBox.Canvas.lineto(2*i,round(LCtr[2]+ p[3]));
  end;

  {complete the curve in LBox to its last point and draw a square around it}
  LBox.canvas.lineto(LRect.right-1,LCtr[2]+round(ipts[tg,cLastPtSpline,3]));
  LBox.Canvas.FillRect(Rect(LRect.right-4,LCtr[2]+round(ipts[tg,cLastPtSpline,3]-3),
       LRect.right+2,LCtr[2]+round(ipts[tg,cLastPtSpline,3]+3)));

  {draw a circle around the first point}
  LBox.Canvas.Ellipse(-4,LCtr[2]+round(ipts[tg,0,3]-4),4,LCtr[2]+round(ipts[tg,0,3]+4));

  {frame square around the last control point}
  LBox.canvas.Brush.Color:= clBlack;
  LBox.Canvas.moveto(LRect.right-4,round(LCtr[2]+ipts[tg,cLastPtSpline,3]-3));
  LBox.Canvas.lineto(LRect.right-4,round(LCtr[2]+ipts[tg,cLastPtSpline,3]+3));
  LBox.Canvas.lineto(LRect.right+2,round(LCtr[2]+ipts[tg,cLastPtSpline,3]+3));
  LBox.Canvas.lineto(LRect.right+2,round(LCtr[2]+ipts[tg,cLastPtSpline,3]-3));
  LBox.Canvas.lineto(LRect.right-4,round(LCtr[2]+ipts[tg,cLastPtSpline,3]-3));

  {Draw a little circle around each of the other control points in LBox}
  for i:= 1 to cLastPtSpline-1 do begin
    absc:= (i*LRect.right) div cLastPtSpline;
    LBox.Canvas.Ellipse(absc - 3,LCtr[2]+round(ipts[tg,i,3]-3),
       absc + 3,LCtr[2]+round(ipts[tg,i,3]+3));
  end;
end;

procedure TGammaPicker.DrawWheel;
{Draws the wheel of hue and saturation. The drawing takes account of the
 angle that the wheel may have been rotated by the user}
  const
    cMedGray= $007F7F7F;
  var
    i,j, k: Integer;
    r: Double;
    hs: Vector2;
    line: pTColorArray;
begin
  for j:= 0 to HSRect.Bottom-1 do begin
    line:= HSBm.ScanLine[j];
    for i:= 0 to HSRect.right-1 do begin
      if PtInWheel(r,i,j) then begin
        if r = 0 then begin {draw the wheel center in grey}
          k:= round(0.55*cMaxComp) and $FF;
          line[i]:= k + k shl 8 + k shl 16;
        end
        else begin
          hs:= FromWheelToHS(i,j); {compute hue and saturation components}
          line[i]:= FromHSLToRGB(hs,0.55); {convert to RGB and draw pixel}
        end
      end;
    end;
  end;
end;

function TGammaPicker.EvaluateBSpline(t: Double): Vector3;
{Compute the value of the coordinates of the point alpha(t) in
 the HSBox (two first coordinates of 'result') and the ordinate
 of the corresponding point of the curve in the LBox. See the
 memory "AproxFuncBSplinesEnglish.pdf"}
var
  a, a0: Double;
  i, k, r, h: Integer;
  p: array[0..3, 0..3] of Vector3;
begin
  a0:= t*cLastPtSpline;
  if a0 < 0 then
    k:= 0
  else if (a0 > cLastPtSpline) then
    k:= cLastPtSpline
  else
    k:= trunc(a0);
  a0:= a0 + 3 - k;
  k:= k+cQ+1;
  for i:= 0 to 3 do
    p[0,i]:= points[0,i+k-3];
  for r:= 1 to 3 do
    for i:= r to 3 do begin
      a:= (a0-i)/(4-r);
      for h:= 1 to 3 do
        p[r,i,h]:= a*p[r-1,i,h]+(1-a)*p[r-1,i-1,h];
    end;
  for i:= 1 to 3 do
    result[i]:= round(p[3,3,i]);
  if result[3] > cMaxLAmplitude then
    result[3]:= cMaxLAmplitude
  else if result[3] < -cMaxLAmplitude then
    result[3]:= -cMaxLAmplitude;
end;



function TGammaPicker.FromHSLToRGB(hs: Vector2; b: Double): TColor;
{it is supposed that both components of 'hs' have absolute value less or
 equal to 1. Then, it is applied a standard formula that may be found in
 the internet}
  const
    oneover3= 1/3;
    oneOver6= 1/6;
    twoOver3= 2/3;
  var
    k: Integer;
    p, q: Double;
    t: Vector3;
begin
  if (b <= 0.5) then
    q:= b*(1 + hs[2])*cMaxComp
  else
    q:= (b + hs[2]*(1-b))*cMaxComp; 
  p:= 2*b*cMaxComp - q;
  t[2]:= hs[1];
  t[1]:= hs[1]+ oneover3;
  if  t[1] > 1 then
    t[1]:= t[1] - 1;
  t[3]:= hs[1]- oneover3;
  if  t[3] < 0 then
    t[3]:= t[3] + 1;
  for k:= 1 to 3 do
    if t[k] < oneOver6 then
      t[k]:= p + 6*(q-p)*t[k]
    else if t[k] < 0.5 then
      t[k]:=q
    else if t[k] < twoOver3 then
      t[k]:= p + 6*(q-p)*(twoOver3- t[k])
    else
      t[k]:= p;
  result:= (round(t[1]) and $FF) +
                     ((round(t[2]) and $FF) shl 8)+
                     ((round(t[3]) and $FF) shl 16);
end;

function TGammaPicker.FromWheelToHS(x,y: Double): Vector2;
{compute the hue and saturation of the color from the coordinates
 in the HSBox taking into account the rotation that the user may
 have done to the wheel. We apply a standard formula obtained in
 internet}
var
  a,b: Double;
begin
  a := HSctr[1] - x;
  b := y - HSctr[2];
  result[2]:= hypot(a,b)/cWheelRadius; {saturation is thus equal to the
     relative farness of the point from the wheel center}
  if result[2] > 0 then begin
    result[2]:= Min(1,result[2]);
    result[1]:= 0.5 + (arctan2(b,a)-phi[tg]) /cTwoPi;
    result[1]:= result[1] - int(result[1]);
    if result[1] > 1 then
      result[1]:= result[1]-1
    else if result[1] < 0 then
      result[1]:= result[1]+1;
  end
  else
    result[1]:= 0.5;
end;


procedure TGammaPicker.Initiate(gtag: Integer; notify: Boolean);
{Draw everything in the tab corresponding to 'gtag' and notify
 the update if it is required}
  var
    i: Integer;
begin
  tg:= gtag;
  for i:= 0 to cLastPtSpline do
    trackingPts[i]:= False;
  DrawWheel;
  PaintHSBox(nil);
  ComputeGamma(tg);
  LBox.Canvas.Draw(0,0,LBm);
  DrawBSplineCurveHSandL;
  if notify then
    DispatchGammaChange;
  trackingHS:= False;
  trackingL:= False;
end;


procedure TGammaPicker.MouseDownInHSBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  var
    c: TColor;
    i: Integer;
begin
  translating:= False;
  spinning:= False;
  if (ssLeft in Shift) then begin
    if (ssCtrl in Shift) then begin {check that the point is inside the
      HS wheel to begin tracking its movements}
      translating:= hypot(x- HSCtr[1],y - HSCtr[2]) < cWheelRadius;
      lastHSPt:= Point(x,y);
    end
    else if ssAlt in Shift then begin
      {the user wants to change the background of the surface drawings}
      c:= HSBm.Canvas.Pixels[x,y]; {this is the selected background color}

      {invert the color components to pass it directly via scanline}
      gammas[tg,0]:= (c and $00FF0000) shr 16 + (c and $0000FF00)+ (c and $000000FF) shl 16;

      if tg < 2 then {both tabs corresponds to the same color scheme, so that
          the background color is always the same}
        gammas[1-tg,0]:= gammas[tg,0];
      DispatchBackgroundChange; {notify the change}
    end
    else if ssShift in Shift then begin
      {the user wants a monochromatic gamma. Colapse the HS curve to a point}
      if hypot(x- HSCtr[1],y - HSCtr[2]) < cWheelRadius then begin
        for i:= 0 to cLastPtSpline do begin {all control points are made equal}
          ipts[tg,i,1]:= x;
          ipts[tg,i,2]:= y;
        end;
        {Recalculate the gamma and draw everything}
        HSBox.Canvas.Draw(0,0,HSBm);
        ComputeGamma(tg);
        DrawBSplineCurveHSandL;
        DispatchGammaChange; {notify the changes}
        lastHSPt:= Point(x,y);
      end
    end
    else begin {the user wants to rotate the weel, or to move a control point}
      spinning:= True;
      lastHSPt:= Point(x,y);

      {check whether the mouse is near some of the control points. Interpret
       the affirmative as if the user wants to move that control}
      for i:= 0 to cLastPtSpline do begin
        trackingPts[i]:=(sqrt(sqr(x-ipts[tg,i,1])+sqr(y-ipts[tg,i,2])) < 5);
        if trackingPts[i] then begin
          spinning:= False;
          break;
        end;
      end;
    end;
  end
end;

procedure TGammaPicker.MouseDownInLBox(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{See doc for Mouse movements in HSBox. Now things are more limited: only
 background changes or vertical movements of the control points are produced}
  var
    i, absc: Integer;
    c: TColor;
begin
    if (ssLeft in Shift) then
      for i:= 0 to cLastPtSpline do begin
        absc:= (i*LRect.right) div cLastPtSpline;
        trackingPts[i]:=(sqrt(sqr(x-absc)+sqr(Lctr[2]+ipts[tg,i,3]-y)) < 5);
        if trackingPts[i] then
          break;
      end;
    if ssAlt in Shift then begin
      c:= LBm.Canvas.Pixels[x,y];
      gammas[tg,0]:= (c and $00FF0000) shr 16 + (c and $0000FF00)+ (c and $000000FF) shl 16;
      DispatchBackgroundChange;
    end;
end;

procedure TGammaPicker.MouseMoveInHSBox(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  var
    i, i0, deltaX, deltaY: Integer;
    moved: Boolean;
    a, b: Double;
begin
  if (not PtInRect(HSRect,Point(x,y))) then
    exit;
  moved:= False;
  if (ssLeft in Shift) then begin
    if translating and (ssCtrl in Shift) then begin
      translating:= hypot(x- HSCtr[1],y - HSCtr[2]) < cWheelRadius;
      if translating then begin
        {user wants to translate the curve in HSBox}
        deltaX := x- lastHSPt.x;
        deltaY := y- lastHSPt.y;
        moved:= True;

        {check that the translation will move the control points inside
        the wheel. Otherwise, do nothing}
        for i:= 0 to cLastPtSpline do
          moved:= moved and ( hypot(ipts[tg,i,1] + deltaX-HSCtr[1],
              ipts[tg,i,2] + deltaY-HSCtr[2]) < cWheelRadius );
        if moved then begin
          for i:= 0 to cLastPtSpline do begin
            ipts[tg,i,1]:= ipts[tg,i,1] + deltaX;
            ipts[tg,i,2]:= ipts[tg,i,2] + deltaY;
          end;
          {Recalculate the gamma and draw everything}
          HSBox.Canvas.Draw(0,0,HSBm);
          ComputeGamma(tg);
          LBox.Canvas.Draw(0,0,LBm);
          DrawBSplineCurveHSandL;
          DispatchGammaChange; {notify the changes}
          lastHSPt:= Point(x,y);
        end;
      end;
    end
    else if spinning  then begin
      {user wants to rotate the wheel}
      a:= hypot(lastHSPt.x- HSCtr[1],lastHSPt.y- HSCtr[2]);
      if a > 1 then begin {otherwise the rotation would be very unstable}
        b:= hypot(x- HSCtr[1],y- HSCtr[2]);
        if b > 1 then begin
          b:= 1/(a*b);
          a:= b*((HSCtr[2]-lastHSPt.y)*(x-HSCtr[1])+(lastHSPt.x- HSCtr[1])*(y-HSCtr[2]));
          {thus, a is the sine of the angle that the mouse has described
           as seen from the wheel center}
          a:= a*(1+ sqr(a)/6); {and this is a good approximation to arcsin(a)}
          phi[tg]:= phi[tg]- a; {change the wheel angle accordingly, draw it
           and everything else}
          DrawWheel;
          HSBox.Canvas.Draw(0,0,HSBm);
          LBox.Canvas.Draw(0,0,LBm);
          DrawBSplineCurveHSandL;
          ComputeGamma(tg);
          DispatchGammaChange;
        end;
      end;
      lastHSPt:= Point(x,y);
    end
    else begin
      {verify that the mouse is really following the same control point}
      for i:= 0 to cLastPtSpline do
        if trackingPts[i] and ((ipts[tg,i,1] <> x) or ((ipts[tg,i,2]<> y)))then begin
          moved:= True;
          i0:= i;
          break;
        end;
      if moved and PtInWheel(a,x,y) then begin
        HSBox.Canvas.Draw(0,0,HSBm); {draw the wheel without the curve}

        {update position of control point}
        ipts[tg,i0,1]:= x;
        ipts[tg,i0,2]:= y;

        {Recalculate the gamma and draw everything}
        ComputeGamma(tg);
        LBox.Canvas.Draw(0,0,LBm);
        DrawBSplineCurveHSandL;
        DispatchGammaChange;
      end;
    end
  end;
end;

procedure TGammaPicker.MouseMoveInLBox(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
{See doc for Mouse movements in HSBox.}
  var
    i, i0: Integer;
    moved: Boolean;
begin
  if (not PtInRect(LRect,Point(x,y))) then
    exit;
  moved:= False;
  if (ssLeft in Shift) then
    for i:= 0 to cLastPtSpline do
      if trackingPts[i] and (Lctr[2]+ipts[tg,i,3]<> y) and
           (abs(y-Lctr[2]) < cMaxLAmplitude) then begin
        moved:= True;
        i0:= i;
        break;
      end;
  if moved then begin
    LBox.Canvas.Draw(0,0,LBm);
    ipts[tg,i0,3]:= y - Lctr[2];
    ComputeGamma(tg);
    DrawBSplineCurveL;
    DispatchGammaChange;
  end;
end;

procedure TGammaPicker.MouseUpInHSBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  var
    i: Integer;
begin
  {initialize trackings to False}
  trackingHS:= False;
  for i:= 0 to cLastPtSpline do
    trackingPts[i]:= False;
end;

procedure TGammaPicker.MouseUpInLBox(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
begin
  for i:= 0 to cLastPtSpline do
    trackingPts[i]:= False;
end;

procedure TGammaPicker.OnChangeGammaTabs(Sender: TObject);
{User has changed the tab in the TGammaPicker. Recalculate and draw the
 new tab}
  var
    ti: Integer;
begin
  if (Sender is TTabControl) then begin
    ti:= GammaTabs.TabIndex;
    if ti < 2 then
      Initiate(ti, false);
  end;
end;

procedure TGammaPicker.OnPaintPicker(Sender: TObject);
begin
  Initiate(tg, true);
end;

procedure TGammaPicker.OnShowGPicker(Sender: TObject);
{position the form in the screen so that be wholly visible}
var
  r: TRect;
begin
  r:= Screen.WorkAreaRect;
  top:= max(r.Top, Top);
  top := min(r.Bottom,top + Height)- Height;
  left := max(r.left,left);
  left:= min(r.right-width, left);
end;

procedure TGammaPicker.PaintHSBox(Sender: TObject);
begin
  HSBox.Canvas.Draw(0,0,HSBm);
  HSBox.Canvas.Pen.Mode:= pmCopy;
  HSBox.Canvas.Pen.Color:= clWhite;
  HSBox.Canvas.Brush.Color := clBlack;
  DrawBSplineCurveHSandL;
end;

procedure TGammaPicker.PaintLBox(Sender: TObject);
begin
  LBox.Canvas.Draw(0,0,LBm);
  LBox.Canvas.Brush.Color := clBlack;
  DrawBSplineCurveL;
end;

function TGammaPicker.PtInWheel(var r: Double; h,v: Integer): Boolean;
{verifies whether the point is inside the wheel}
begin
  r:= hypot(h - HSCtr[1],v-HSCtr[2]);
  result:= ( r <= cWheelRadius ) ;
end;

end.
