{-----------------------------------------------------------------------------
 Unit Name: FImmersionR3R6
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
 Purpose: Implements class 'TImmersionR3R6' which is the main form of the
          application, with the menu
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit FImmersionR3R6;

interface

uses
  SysUtils, Windows, Graphics, Messages, Classes, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, StdActns, ActnList, Menus, Printers, ComCtrls,
  VectorsAndUtil, Grids, ValEdit, ImplicitRenderThread, FMapR3R6;

const                                                                
 {For saving and reading to files}

  kFilter = 'i36'; {goes in InputStrs[0]}
  kCreator = 'Angel Montesinos-Amilibia';  {goes in InputStrs[1]}
  kVersion000 = '0'; {goes in InputStrs[2]}
  kLastStr = 19; {goes in InputStrs[3]. InputStrings[4] is left unused in this
                 version. InputStrs go from 5 to kLastStr. These
                 strings are the data to save.}

 {Colors }

  cWindowColor = $00B0E8FF;


{Miscellaneous}
  kReturn = char(13);
  kEnter = char(3);
  kEscape = char(27);
  kTab = char($09);
  kPunto = '.';

  kGrid = 50;
  kPrintGrid = 80;

type
  
  TImmersionR3R6 = class(TForm)
    x1Edit: TEdit;
    Label1 : TLabel;
    x2Edit: TEdit;
    Label5 : TLabel;
    x3Edit: TEdit;
    Label6 : TLabel;
    x4Edit: TEdit;
    Label7 : TLabel;
    uMinEdit : TEdit;
    uMaxEdit : TEdit;
    Label8 : TLabel;
    vMinEdit : TEdit;
    vMaxEdit : TEdit;
    Label9 : TLabel;
    EnterButton : TSpeedButton;
    MainMenu1 : TMainMenu;
    File1 : TMenuItem;
    Open1 : TMenuItem;
    Save1 : TMenuItem;
    SaveAs1 : TMenuItem;
    Print1 : TMenuItem;
    N1 : TMenuItem;
    Exit1 : TMenuItem;
    Edit1 : TMenuItem;
    Cut1 : TMenuItem;
    Copy1 : TMenuItem;
    Paste1 : TMenuItem;
    EditCut1 : TEditCut;
    EditCopy1 : TEditCopy;
    EditPaste1 : TEditPaste;
    ActionList1 : TActionList;
    OpenDialog1 : TOpenDialog;
    SaveDialog1 : TSaveDialog;
    PrintDialog1 : TPrintDialog;
    MainBox: TPaintBox;
    InfoBar: TStatusBar;
    Manual1: TMenuItem;
    N7: TMenuItem;
    Label10: TLabel;
    x5Edit: TEdit;
    FunctionValues1: TValueListEditor;
    View1: TMenuItem;
    ShowPalette1: TMenuItem;
    Label2: TLabel;
    x6Edit: TEdit;
    Label3: TLabel;
    wMinEdit: TEdit;
    wMaxEdit: TEdit;
    Label4: TLabel;
    uEdit: TEdit;
    Label11: TLabel;
    vEdit: TEdit;
    Label12: TLabel;
    wEdit: TEdit;
    DomainBox: TPaintBox;
    wMeter: TPaintBox;
    N2: TMenuItem;
    Monocolor1: TMenuItem;
    Bicolor1: TMenuItem;
    N3: TMenuItem;
    Veroneseofcurvature1: TMenuItem;
    FocalLocus1: TMenuItem;
    N4: TMenuItem;
    ResetCameras1: TMenuItem;
    Background1: TMenuItem;
    NormofH1: TMenuItem;
    MeanCurvatureEnergy1: TMenuItem;
    ShowTangentSpace1: TMenuItem;

    procedure CompleteGammasStuff1;
    procedure ComputeAndShowData(x: Vector3);
    procedure DrawCursor;
    procedure DrawFocalLocus;
    procedure DrawScene;
    procedure DrawVeronese;
    procedure FillDomainWithFunction;
    function FromDisk: AnsiString;
    procedure ManageMenus;
    procedure HandleThreadDone(sender: TObject);
    procedure OnCloseQuery(Sender : TObject; var CanClose : Boolean);
    procedure OnCreateForm(Sender : TObject);
    procedure OnDestroyForm(Sender: TObject);
    procedure OnEnterButton(Sender : TObject);
    procedure OnExit(Sender : TObject);
    procedure OnMouseDown(Sender : TObject; Button : TMouseButton;
      Shift : TShiftState; X, Y : Integer);
    procedure OnMouseDownInCursor(Sender: TObject; Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer);
    procedure OnMouseDownInDomain(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnMouseMove(Sender : TObject; Shift : TShiftState; X,
      Y : Integer);
    procedure OnMouseMoveInDomain(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OnMouseMoveInWCursor(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure OnMouseUp(Sender : TObject; Button : TMouseButton;
      Shift : TShiftState; X, Y : Integer);
    procedure OnMouseUpDomain(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnMouseUpInCursor(Sender: TObject; Button: TMouseButton;
          Shift: TShiftState; X, Y: Integer);
    procedure OnOpen(Sender : TObject);
    procedure OnPaint(Sender: TObject);
    procedure OnPrint(Sender : TObject);
    procedure OnSave(Sender : TObject);
    procedure OnSaveAs(Sender : TObject);
    function ReadData: Boolean;
    procedure ToDisk;
    procedure OnKeyUpInEditField(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure OnPreferences(Sender: TObject);
    procedure OnPaintDomBox(Sender: TObject);
    procedure OnViewMenu(Sender: TObject);
    procedure OnPaintWCursor(Sender: TObject);
    procedure OnManual(Sender: TObject);
    procedure OnChangeEditField(Sender: TObject);
    procedure StopThreads;
    procedure OnBackground(Sender: TObject);
  private
    domBmp: TBitmap;
    firstPt, prevSpot : TPoint;
    thereIsSpot: Boolean;
    compEdit: array[1..3] of TEdit;
    boundsEdit: array[1..3,1..2] of TEdit;
  public
    InputStrs : array[0..kLastStr] of AnsiString;
    rctMain, rctDomain, rctCursor: TRect;
    min3D, max3D, edges3D, halfEdges3D, center3D, xDom: Vector3;
    grid, thrNumber: Integer;
    span3D, radius3D: Double;
    alreadyParsed, alreadyDrawn, tracking,
         trackingDom, trackWCrs: Boolean;
    printEquations, PrintSubText, printingFocalLocus, dirty: Boolean;
    prTitle, prSubText: AnsiString;
    prResolution, wCrsPos: Integer;
    showVeronese: Boolean;
    Filename : TFilename;
    procedure DrawSpot(pt: TPoint); overload;
    procedure DrawSpot(x: Vector3); overload;
    procedure Wash; 
  end;

var
  ImmersionR3R6 : TImmersionR3R6;
  thrs: array of TImplicitRenderThread; {used in drawing the Focal locus}

implementation

{$R *.DFM}

uses

  Math,
  OCFunctDDDoubleSafe, ResourceStrings,
  PrintingPrefsDlg, FYesNotCancel, FGammaPicker,
  FDrawParamSurf, Printing, FManImmersionR3R6, FRenderImplicitSurfaceWithThreads,
  FSupplWindow;



procedure DefinePrintingRectangle(var destRect: TRect; PageWidth, PageHeight :
  Integer);
{set the rectangle in the printed page where the drawing is to be done}
var
  marH, marV, side: Integer;
begin
  marH := PageWidth div 10;
  marV := PageHeight div 10;
  side := PageWidth - 2 * marH;
  if side > (PageHeight - 5 * marV) then
    side := PageHeight - 5 * marV;
  marH := (PageWidth - side) div 2;
  marV := (pageHeight - side) div 7;
  setRect(destRect, marH, marV, marH + side, marV + side);
end;



//Methods of class TImmersionR3R6


procedure TImmersionR3R6.CompleteGammasStuff1;
{Initialize gPicker with the defaults. Used on creating the main form}
const
    cTwoPi = 6.283185;
var
  i, j: Integer;
  co, si, psi: Single;
begin
  with gPicker do begin
    for i:= 0 to 1 do begin
      phi[i]:= (i+1)* 0.8;
      psi:= cTwoPi/(cLastPtSpline+1);
      for j:= 0 to cLastPtSpline do begin
        co:= 12*cos(j*psi);
        si:= 12*sin(j*psi);
        ipts[i,j,1]:= HSctr[1] + cWheelRadius/2 + co;
        ipts[i,j,2]:= HSctr[2] + cWheelRadius/2 + si;
        ipts[i,j,3]:= 0.8*LRect.bottom*(0.5-j/cLastPtSpline);
      end;
    end;
  end;
end;

procedure TImmersionR3R6.ComputeAndShowData(x: Vector3);
{For showing data in the FunctionValues panel}
var
  Hnorm, MeanCurvEnerg: Double;
  FFRank, Vrank: Integer;
  separator: Char;
begin
  separator := DecimalSeparator; {save decimal separator}
  DecimalSeparator := '.';  {only this decimal separator is admitted}
  if MapR3R6.ComputeVeroneseInvariants(x,FFrank, Vrank,Hnorm,MeanCurvEnerg) then begin
    with FunctionValues1.TitleCaptions do begin {title captions}
      BeginUpdate;
      Clear;
      Append(eFunction);
      Append(eValue);
      EndUpdate;
    end;
    with FunctionValues1.Strings do begin
      Strings[0]:= 'u=' + FloatToStrF(x[1], ffFixed, 14, 10);
      Strings[1]:= 'v=' + FloatToStrF(x[2], ffFixed, 14, 10);
      Strings[2]:= 'w=' + FloatToStrF(x[3], ffFixed, 14, 10);
      Strings[3]:= eFFRank + IntToStr(FFrank);
      Strings[4]:= eVeroneseRank + IntToStr(Vrank);
      Strings[5]:= eNormH + FloatToStrF(Hnorm, ffFixed, 14, 10);
      Strings[6]:= eMeanCurvEnergyEq + FloatToStrF(MeanCurvEnerg, ffFixed, 14, 10);
    end;
    FunctionValues1.Update;
  end;
  DecimalSeparator := separator; {restore decimal separator}
end;

procedure TImmersionR3R6.DrawCursor;
{Draw the cursor in the w-meter}
begin
  with wMeter.Canvas do begin
    FillRect(rctCursor);
    Pen.Color:= $00FFCCCC;
    Pen.Width:= 2;
    Pen.Mode := pmCopy;
    moveto(0,6);
    lineto(kCrsWidth,6);
    Pen.Color:= clGray;
    moveto(wCrsPos,1);
    lineto(wCrsPos,12);
  end;
end;

procedure TImmersionR3R6.DrawFocalLocus;
var
  ms: AnsiString;
begin
  if MapR3R6.ComputeVeroneseData(xDom, grid) then
    ms:= rist.ComputeAndDraw(kWidthMainBox, kWidthMainBox, True);
end;

procedure TImmersionR3R6.DrawScene;
begin
  if showVeronese then begin
    DrawVeronese;
    MainBox.canvas.Draw(0,0,bmp);
  end
  else
    DrawFocalLocus;
end;

procedure TImmersionR3R6.DrawVeronese;
var
  min2D, max2D: Vector2;
  i: Integer;
begin
  if MapR3R6.ComputeVeroneseData(xDom, grid) then begin
    dps.ReceiveBasicData(mapR3R6.min3DNorm, mapR3R6.max3DNorm, grid, MainBox.Width);
    dps.ComputeSurface;
    alreadyDrawn:= True;
    SupplWindow.DrawContents;
  end;
end;

procedure TImmersionR3R6.FillDomainWithFunction;
{Fill the domain panel with a color that depends on the
value of a function selected by user in the menu}
var
  fVal: Double;
  x: Vector3;
  i, j: Integer;
  line: pTColorArray;
  c: TColor;
begin
  c:= gammas[0,0];
  x[3]:= xDom[3];
  for j:= 0 to kWidthDomBox-1 do begin
    line:= domBmp.scanline[j];
    x[2]:= max3D[2]  - j*edges3D[2]/kWidthDomBox;
    for i:= 0 to kWidthDomBox-1 do begin
      x[1]:= min3D[1] + i*edges3D[1]/kWidthDomBox;
      if MapR3R6.backF(x,fVal) then {convert function value to color, and
           fill the pixel with it}
        line^[i]:= bckgPal[round( cGammaN*( 0.5 + arctan(fVal)/pi))]
      else
        line^[i]:= c;
    end;
  end;
end;

procedure TImmersionR3R6.DrawSpot(pt: TPoint);
{draw in the Domain panel the little circle that marks the relative
 position of the current first two coordinates value}
var
  oldColor: TColor;
begin
  with DomainBox.canvas do begin
    oldColor:= Pen.Color;
    Pen.Color:= clBlack;
    Pen.Mode:= pmNotXor;
    if thereIsSpot then begin
      {draw the circle in the previous position. Using the pmNotXoR
      mode, this erases the previous circle}
      moveto(prevSpot.x - 4, prevSpot.y + 1);
      lineto(prevSpot.x - 2, prevSpot.y + 3);
      lineto(prevSpot.x + 1, prevSpot.y + 3);
      lineto(prevSpot.x + 3, prevSpot.y + 1);
      lineto(prevSpot.x + 3, prevSpot.y - 2);
      lineto(prevSpot.x + 1, prevSpot.y - 4);
      lineto(prevSpot.x - 2, prevSpot.y - 4);
      lineto(prevSpot.x - 4, prevSpot.y - 2);
      lineto(prevSpot.x - 4, prevSpot.y + 1);
      thereIsSpot:= False;
    end;

    {draw the circle in the current position. Using the pmNotXoR
     mode, this allows to erase the circle later}
    moveto(pt.x - 4, pt.y + 1);
    lineto(pt.x - 2, pt.y + 3);
    lineto(pt.x + 1, pt.y + 3);
    lineto(pt.x + 3, pt.y + 1);
    lineto(pt.x + 3, pt.y - 2);
    lineto(pt.x + 1, pt.y - 4);
    lineto(pt.x - 2, pt.y - 4);
    lineto(pt.x - 4, pt.y - 2);
    lineto(pt.x - 4, pt.y + 1);
    prevSpot:= pt;
    thereIsSpot:=  True;
    Pen.Color := oldColor;
    Pen.Mode := pmCopy;
  end;
end;

procedure TImmersionR3R6.DrawSpot(x: Vector3);
{the same with other calling parameters}
var
  pt: TPoint;
begin
  pt.x:= round((x[1]-min3D[1])*kWidthDomBox/edges3D[1]);
  pt.y:= round((max3D[2]-x[2])*kWidthDomBox/edges3D[2]);
  DrawSpot(pt);
end;


function TImmersionR3R6.FromDisk: AnsiString;
{read data from disk and fill the corresponding edit fields}
var
  sl: TStringList;
begin
  result:= '';
  sl:= TStringList.Create;
  sl.LoadFromFile(fileName);
  try
    if (sl.Strings[0] <> kFilter) or (sl.Strings[1] <> kCreator)  then
      raise Exception.Create(eWrongFile);
    if (sl.Strings[2] = kVersion000) then begin
      if StrToInt(sl.Strings[3]) <> kLastStr then
        raise Exception.Create(eCorruptedFile);
      {in version kVersion000, StrToInt(sl.Strings[4] is purposely left unused}
      x1Edit.Text:= sl.Strings[5];
      x2Edit.Text:= sl.Strings[6];
      x3Edit.Text:= sl.Strings[7];
      x4Edit.Text:= sl.Strings[8];
      x5Edit.Text:= sl.Strings[9];
      x6Edit.Text:= sl.Strings[10];
      uMinEdit.Text:= sl.Strings[11];
      uMaxEdit.Text:= sl.Strings[12];
      vMinEdit.Text:= sl.Strings[13];
      vMaxEdit.Text:= sl.Strings[14];
      wMinEdit.Text:= sl.Strings[15];
      wMaxEdit.Text:= sl.Strings[16];
      uEdit.Text:= sl.Strings[17];
      vEdit.Text:= sl.Strings[18];
      wEdit.Text:= sl.Strings[19];
    end
  finally
    FreeAndNil(sl);
  end;
end;

procedure TImmersionR3R6.HandleThreadDone(sender: TObject);
{A thread for drawing the Focal locus has terminated. If no other threads
 are running, draw Focal locus to window, or continue printing}
 var
   x,y: Vector3;
   i: Integer;
begin
  dec(thrNumber);
  if thrNumber <= 0  then begin
    if not printingFocalLocus then begin
      alreadyDrawn:= True;

      {draw the mean curvature vector}
      x[1]:= 0;
      x[2]:= 0;
      x[3]:= 0;
      for i:= 1 to 3 do
        y[i]:= -MapR3R6.meanCurv[i];
      rist.DrawSegment(bmp.canvas, x,y,5,clLime);
      MainBox.Canvas.Draw(0,0, bmp);
      rist.busy:= False;
      SupplWindow.DrawContents;
    end
    else begin
      PrintAll;
      printingFocalLocus:= False;
    end;
  end;
end;


procedure TImmersionR3R6.ManageMenus;
{update menus}
begin
  Save1.Enabled := alreadyParsed and dirty;
  SaveAs1.Enabled := alreadyParsed;
  Print1.Enabled := alreadyParsed;
end;

procedure TImmersionR3R6.OnBackground(Sender: TObject);
{color the domain panel according to the values of the selected function}
begin
  case TMenuItem(sender).tag of
    0: MapR3R6.backF:= MapR3R6.NormOfH;
    1: MapR3R6.backF:= MapR3R6.MeanCurvEnergy;
  else
    ;
  end;
  if alreadyDrawn and not rist.busy then begin
    FillDomainWithFunction;
    domainBox.invalidate;
  end;
end;

procedure TImmersionR3R6.OnCloseQuery(Sender : TObject; var CanClose : Boolean);
{called when the user closes the main window}
var
  res: Word;
begin
  if dirty then begin
    with TYesNotCancel.Create(nil) do  {give user option of saving data}
      try
        res:= ShowModal;
      finally
        Release;
      end;
    if res = mrNo then
      CanClose := True
    else if res = mrYes then begin
      OnSave(sender);   {user opted for saving data}
      CanClose := not dirty;
    end
    else
      CanClose := False; {user cancelled}
  end
  else
    CanClose := True;
end;


procedure TImmersionR3R6.OnCreateForm(Sender : TObject);
var
  i: Integer;
begin
  color:= cWindowColor;
  FunctionValues1.FixedColor:= cWindowColor;
  try
    {Items in File menu}
    File1.Caption:= eFile;
    Manual1.Caption:= eManualDots;
    Open1.caption:= eOpen;
    Save1.caption:= eSave;
    SaveAs1.caption:= eSaveAs;
    Print1.caption:= ePrint;
    Exit1.caption:= eExit;

    {Items in Edit menu}
    Edit1.Caption:= eEdit;
    Cut1.caption:= eCut;
    Copy1.caption:= eCopy;
    Paste1.caption:= ePaste;

    {Items in View menu}
    View1.caption:= eView1;
    ShowPalette1.caption:= eShowPalette;
    ShowTangentSpace1.caption:= eShowTangentSpace;
    Monocolor1.caption:= eMonocolor;
    Bicolor1.caption:= eBicolor;
    VeroneseOfCurvature1.caption:= eVeroneseCurv;
    FocalLocus1.caption:= eFocalLocus;
    ResetCameras1.caption:= eResetCameras;

   {Items in Background menu}
    NormOfH1.caption:= eNormOfH;
    MeanCurvatureEnergy1.caption:= eMeanCurvEnergy;
    
    {strings in FunctionValues1}
    with FunctionValues1.TitleCaptions do begin
      BeginUpdate;
      Clear;
      Append(eFunction);
      Append(eValue);
      EndUpdate;
    end;
    with FunctionValues1.Strings do begin
      Strings[0]:= 'u=';
      Strings[1]:= 'v=';
      Strings[2]:= 'w=';
      Strings[3]:= eFFrank;
      Strings[4]:= eVeroneseRank;
      Strings[5]:= eNormH;
      Strings[6]:= eMeanCurvEnergyEq;
    end;
    FunctionValues1.Update;

    rctMain:= Rect(0,0,kWidthMainBox,kHeightMainBox);
    rctDomain:= Rect(0,0,kWidthDomBox,kHeightDomBox);
    rctCursor:= Rect(0,0,wMeter.Width, wMeter.Height);
    max3D[3]:= StrToFloat(wMaxEdit.text);
    min3D[3]:= StrToFloat(wMinEdit.text);
    wCrsPos:= wMeter.Width div 2;
    mainBox.canvas.FillRect(rctMain);
    thereIsSpot:= False;
    alreadyParsed:= False;
    tracking:= False;
    trackingDom:= False;
    showVeronese:= True;
    grid:= kGrid;

    {initialize some strings that will being saved to all files
     created by this program}
    inputStrs[0]:= kFilter;
    inputStrs[1]:= kCreator;
    inputStrs[2]:= kVersion000;
    inputStrs[3]:= IntToStr(kLastStr);
    inputStrs[4]:= ' '; {left unused in version kVersion000}

    compEdit[1]:= uEdit;
    compEdit[2]:= vEdit;
    compEdit[3]:= wEdit;
    boundsEdit[1,1]:= uMinEdit;
    boundsEdit[1,2]:= uMaxEdit;
    boundsEdit[2,1]:= vMinEdit;
    boundsEdit[2,2]:= vMaxEdit;
    boundsEdit[3,1]:= wMinEdit;
    boundsEdit[3,2]:= wMaxEdit;

    {we give these value so that nothing will appear selected in the box}
    FunctionValues1.Selection:= TGridRect(Rect(20,20,20,20));
    MapR3R6:= TMapR3R6.Create;

   {Create and initialize color pickers}
    gPicker:= TGammaPicker.Create(self);
    CompleteGammasStuff1;
    for i:= 0 to cLastGamma do
      gPicker.ComputeGamma(i);
    for i:= cLastGamma downto 0 do begin
      gPicker.Initiate(i,false);
      gammas[i,0]:= $00DDDDFF;
    end;
    gPicker.Show;

    domBmp:= TBitmap.Create;
    domBmp.Width:= kWidthDomBox;
    domBmp.Height:= kWidthDomBox;
    domBmp.PixelFormat := pf32bit;
    domBmp.Canvas.Brush.Color := clWhite;
    domBmp.Canvas.Brush.Style := bsSolid;
    SupplWindow:= TSupplWindow.Create(self);
  except on Exception do begin
    OnDestroyForm(nil);
    Halt;
    end;
  end;
end;

procedure TImmersionR3R6.OnDestroyForm(Sender: TObject);
begin
  FreeAndNil(gPicker);
  FreeAndNil(MapR3R6);
  thrs:=  nil;
  domBmp.Free;
end;

procedure TImmersionR3R6.OnEnterButton(Sender : TObject);
var
  i: Integer;
begin
  if rist.busy and (high(thrs)>=0) then
    for i:= 0 to high(thrs) do
      thrs[i].Terminate;
  if ReadData then begin
    {we begin always drawing first the Veronese surface of the new immersion}
    ShowVeronese:= True;
    VeroneseOfCurvature1.Checked:= True;
    FocalLocus1.Checked:= False;
    dps.InitializeCameras(kWidthMainBox);
    rist.InitializeCubeViewpointAndFocus;
    DrawVeronese;
    MainBox.canvas.Draw(0,0,bmp);
    FillDomainWithFunction;
    DomainBox.canvas.Draw(0,0,domBmp);
    thereIsSpot:= False;
    DrawSpot(xDom);
    wCrsPos:= round((xDom[3]-min3D[3])*kCrsWidth/edges3D[3]);
    DrawCursor;
  end;
end;



procedure TImmersionR3R6.OnExit(Sender : TObject);
begin
  if rist.busy then
    Exit;
  OnSave(sender);
  Application.Terminate;
end;


procedure TImmersionR3R6.OnKeyUpInEditField(Sender: TObject; var Key: Word;
  Shift: TShiftState);
{Is, as in Mathematica, user clicks <shift>+<return>, the program reads
 the input}
begin
  if rist.busy then
    Exit;
  if (Key = VK_RETURN) and (ssShift in Shift) then
    OnEnterButton(nil);
end;

procedure TImmersionR3R6.OnMouseDown(Sender : TObject; Button : TMouseButton;
  Shift : TShiftState; X, Y : Integer);
{mouse was clicked in the main panel}
begin
  if rist.busy then
    Exit;
  if alreadyParsed then begin
    InfoBar.SimpleText := '';
    firstPt.X := X;
    firstPt.Y := Y;
    tracking:=  (ssRight in Shift);
  end;
end;

procedure TImmersionR3R6.OnMouseDownInCursor(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{user begins to move the w-cursor}
begin
  if rist.busy then
    Exit;
  if sender is TPaintBox then
    trackWCrs:= ( abs(x-wCrsPos) < 4);
end;

procedure TImmersionR3R6.OnMouseMove(Sender : TObject; Shift : TShiftState; X,
  Y : Integer);
{mouse is moving on the main panel. Move surface accordingly}
begin
  if rist.busy then
    Exit;
  if alreadyParsed and tracking then begin
    if showVeronese then begin
      dps.MoveSurface(shift, firstPt,Point(x,y));
      {dps.DrawCurve(bmp.canvas,mapR3R6.SecondFF, asymptLocus);}
    end
    else begin
      rist.MoveSurface(shift, firstPt,Point(x,y));
      firstPt:= Point(x,y);
      Exit;
    end;
    MainBox.Canvas.Draw(0,0,bmp);
    firstPt:= Point(x,y);
  end;
end;

procedure TImmersionR3R6.OnMouseMoveInWCursor(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  xx: Vector3;
  ds: char;
begin
  if (rist.busy) or not alreadyParsed then
    Exit;
  if (ssleft in Shift) then begin {update scene}
    if (x > 0) and (x < kCrsWidth) and(x <> wCrsPos) then begin
      wCrsPos:= x;
      xDom[3]:= min3D[3] + x*edges3D[3]/kCrsWidth;
      ds:= DecimalSeparator;
      DecimalSeparator:= '.';
      wEdit.text:= FormatFloatString(xDom[3], 8);
      DecimalSeparator:= ds;
      wEdit.Update;
      wEdit.modified:= False;  {with this trick, wEdit will not fire an
                                'onChange' event}
      DrawCursor;
      DrawScene;
      dirty:= True;
    end;
  end
  else begin {only show data}
    xx[1]:= xDom[1];
    xx[2]:= xDom[2];
    xx[3]:= min3D[3] + x*edges3D[3]/kCrsWidth;
    ComputeAndShowData(xx);
  end;
end;

procedure TImmersionR3R6.OnMouseUp(Sender : TObject; Button : TMouseButton;
  Shift : TShiftState; X, Y : Integer);
begin
  tracking:= False;
end;

procedure TImmersionR3R6.OnMouseUpInCursor(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{Since the computation of the colors at each pixel of the domain is very
 expensive in CPU time, we do it only after the mouse has been released}
begin
  trackWCrs:= False;
  FillDomainWithFunction;
  DomainBox.canvas.Draw(0,0,domBmp);
  thereIsSpot:= False;
  DrawSpot(prevSpot);
end;

procedure TImmersionR3R6.OnOpen(Sender : TObject);
{open a file of the correct extension and input the data}
var
  str: AnsiString;
  Done : Boolean;
begin
  if rist.busy then
    Exit;
  done := False;
  OpenDialog1.Options := [ofFileMustExist, ofHideReadOnly, ofNoChangeDir];
  while not Done do
    begin
      if OpenDialog1.Execute then
        begin
          if not (ofExtensionDifferent in OpenDialog1.Options) then
            begin
              fileName := OpenDialog1.FileName;
              FromDisk; {fill the edit fields with data from file}
              if ReadData then begin {read data as if user had written it to window}
                VeroneseOfCurvature1.Checked:= True;
                FocalLocus1.Checked:= False;
                DrawVeronese;
              end;
              mainBox.Invalidate;
              str:= ExtractFilename(filename);
              caption:= copy(str,1, length(caption)-length(kFilter)-1);
              dirty := False;
              ManageMenus;
              Done := True;
            end
        end
      else { User cancelled }
        Done := True;
    end;
end;

procedure TImmersionR3R6.OnPaint(Sender: TObject);
{the panel need redrawing }
begin
  MainBox.Canvas.Draw(0, 0, bmp);
end;

procedure TImmersionR3R6.OnPrint(Sender : TObject);
var
  rslt: integer;
  answer: Byte;
begin
  if (rist.busy) or (not alreadyParsed) then
    Exit;
  answer:= mrOK;
  rslt := Form2.ShowModal; {open dialog for setting some preferences}
  if (rslt = mrOK) then begin
    if PrintDialog1.Execute then
      update
    else
      answer:= mrCancel;
    if answer = mrOk then begin
      Printer.orientation := poPortrait;
      Printer.BeginDoc;
      answer:= PreparePrint; {'PreparePrint' will call 'PrintAll' that
         will call 'Printer.EndDoc'}
      if answer<>mrOK then
        InfoBar.SimpleText:= eErrorInPrinting
    end;
  end
  else
    InfoBar.SimpleText:= ePrintingCancelled;
end;


procedure TImmersionR3R6.OnSave(Sender : TObject);
begin
  if rist.busy then
    Exit;
  if alreadyParsed and dirty then begin
    if (filename = '') then
      OnSaveAs(sender)
    else begin
      ToDisk;
      dirty := False;
    end;
  end;
end;

procedure TImmersionR3R6.OnSaveAs(Sender : TObject);
var
  Done : Boolean;
begin
  if rist.busy then
    Exit;
  if alreadyParsed then begin
    Done := False;
    while not Done do begin
      if SaveDialog1.Execute then
        begin
          filename := SaveDialog1.FileName;
          caption:= ExtractFilename(filename);
          caption:= copy(caption,1, length(caption)-length(kFilter)-1);
          ToDisk;
          dirty := False;
          Done := True;
        end
      else { User cancelled }
        Done := True;
    end;
  end;
end;



function TImmersionR3R6.ReadData: Boolean;
{read the input written by user or extracted from a file}
var
  varText, m: ansiString;
  i : Integer;
begin
  result:= True;
  InfoBar.SimpleText := '';
  m := eErrorReadingEditField; {default message}
  try
    varText := Trim(x1Edit.Text);
    x1Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w'  and no parameters}
    m:= theFs[1].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[1].BuildFirstDerivative;
    theFs[1].BuildSecondDerivative;
    inputStrs[5]:= varText;

    varText := Trim(x2Edit.Text);
    x2Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w' and no parameters}
    m:= theFs[2].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[2].BuildFirstDerivative;
    theFs[2].BuildSecondDerivative;
    inputStrs[6]:= varText;

    varText := Trim(x3Edit.Text);
    x3Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w' and no parameters}
    m:= theFs[3].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[3].BuildFirstDerivative;
    theFs[3].BuildSecondDerivative;
    inputStrs[7]:= varText;

    varText := Trim(x4Edit.Text);
    x4Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w' and no parameters}
    m:= theFs[4].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[4].BuildFirstDerivative;
    theFs[4].BuildSecondDerivative;
    inputStrs[8]:= varText;

    varText := Trim(x5Edit.Text);
    x5Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w' and no parameters}
    m:= theFs[5].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[5].BuildFirstDerivative;
    theFs[5].BuildSecondDerivative;
    inputStrs[9]:= varText;

    varText := Trim(x6Edit.Text);
    x6Edit.Text := varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    {build a function with variables 'u,v,w' and no parameters}
    m:= theFs[6].ParseNewText(varText, 'uvw','');
    if m<>'' then
      raise Exception.Create(m);
    theFs[6].BuildFirstDerivative;
    theFs[6].BuildSecondDerivative;
    inputStrs[10]:= varText;

    varText := Trim(uMinEdit.Text);
    uMinEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, min3D[1]);
      inputStrs[11]:= varText;

    varText := Trim(uMaxEdit.Text);
    uMaxEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, max3D[1]);
      inputStrs[12]:= varText;
    if max3D[1] <= min3D[1] then
      raise Exception.Create(euMinLessThanuMax);

    varText := Trim(vMinEdit.Text);
    vMinEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, min3D[2]);
      inputStrs[13]:= varText;

    varText := Trim(vMaxEdit.Text);
    vMaxEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, max3D[2]);
      inputStrs[14]:= varText;
    if max3D[2] <= min3D[2] then
      raise Exception.Create(evMinLessThanvMax);

    varText := Trim(wMinEdit.Text);
    wMinEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, min3D[3]);
      inputStrs[15]:= varText;

    varText := Trim(wMaxEdit.Text);
    wMaxEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, max3D[3]);
      inputStrs[16]:= varText;
    if max3D[3] <= min3D[3] then
      raise Exception.Create(ewMinLessThanwMax);

    varText := Trim(uEdit.Text);
    uEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, xDom[1]);
      inputStrs[17]:= varText;

    varText := Trim(vEdit.Text);
    vEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, xDom[2]);
      inputStrs[18]:= varText;

    varText := Trim(wEdit.Text);
    wEdit.Text:= varText;
    if (varText = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varText, xDom[3]);
      inputStrs[19]:= varText;

    {Set some basic data }
    for i:= 1 to 3 do begin
      edges3D[i] := Max3D[i] - Min3D[i];
      halfEdges3D[i] := edges3D[i] / 2;
      center3D[i] := (Max3D[i] + Min3D[i]) / 2;
    end;
    span3D := EuclNorm(edges3D);
    radius3D := span3D / (kWidthMainBox + kHeightMainBox);
    dirty := True;
    alreadyParsed:= True;

    {erase everything visible}
    mainBox.canvas.FillRect(rctMain);
    InfoBar.SimpleText:='';
    ManageMenus;
  Except on E: Exception do begin
      Beep(100,100);
      InfoBar.SimpleText := m;
      AlreadyParsed:= False;
      result:= False;
    end;
  end;
end;




procedure TImmersionR3R6.ToDisk;
{save to a file the strings that define the surface}
var
  sl: TStringList;
  i: Integer;
begin
  sl:= TStringList.Create;
  try
    sl.BeginUpdate;
    sl.Clear;
    {$O-}  {It is mandatory that the list be built in the increasing order
            of 'i'. So, avoid bad tricks by the compiler optimizer by
            desactivating it. I learned this lesson in the hard way}
    for i:= 0 to kLastStr do
      sl.Append(inputStrs[i]);
    sl.EndUpdate;
    {$O+}  {restore compiler optimizing}
    sl.SaveToFile(fileName);
  finally
    FreeAndNil(sl);
  end;
end;



procedure TImmersionR3R6.OnPreferences(Sender: TObject);
begin
  if Sender = ShowPalette1 then
    gPicker.Show;
end;

procedure TImmersionR3R6.OnPaintDomBox(Sender: TObject);
begin
  if alreadyParsed then begin
    DomainBox.canvas.Draw(0,0,domBmp);
    thereIsSpot:= False;
    DrawSpot(xDom);
  end
  else
    DomainBox.canvas.FillRect(rctDomain);
end;

procedure TImmersionR3R6.OnMouseDownInDomain(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if sender is TPaintBox then begin
    trackWCrs:= ( abs(x-wCrsPos) < 4);
  end
end;

procedure TImmersionR3R6.OnMouseMoveInDomain(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  xx: Vector3;
  ds: char;
begin
  if rist.busy then
    Exit;
  if alreadyParsed then begin
    ds:= DecimalSeparator;
    DecimalSeparator:= '.';
    {change parameter values}
    xx[1]:= min3D[1] + x*edges3D[1]/kWidthDomBox;
    xx[2]:= max3D[2] - y*edges3D[2]/kHeightDomBox;
    xx[3]:= xDom[3];

    ComputeAndShowData(xx);
    if (ssLeft in Shift) then begin
      {change the scene and the corresponding edit fields}
      xDom:= xx;
      DrawSpot(Point(x,y));
      DrawScene;
      uEdit.text := FloatToStrF(xDom[1], ffFixed, 4, 4);
      vEdit.text := FloatToStrF(xDom[2], ffFixed, 4, 4);
      uEdit.update;
      uEdit.Modified:= False;
      vEdit.update;
      vEdit.Modified:= False;
    end;
    DecimalSeparator:= ds;
  end;
end;

procedure TImmersionR3R6.OnMouseUpDomain(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  trackingDom:= False;
end;

procedure TImmersionR3R6.OnViewMenu(Sender: TObject);
var
  clrs: Byte;
  showVer: Boolean;
begin
  If not alreadyDrawn then
    Exit;
  showVer:= ShowVeronese;
  clrs:= dps.numberOfColors;
  if sender  =  ShowPalette1 then
    gPicker.Show
  else if sender = ShowTangentSpace1 then
    SupplWindow.Show
  else if sender = VeroneseOfCurvature1 then
    showVer:= True
  else if sender =  FocalLocus1 then
    showVer:= False
  else if sender = Monocolor1 then
    clrs:= 0
  else if sender = Bicolor1 then
    clrs:= 1
  else if sender = ResetCameras1 then begin
    if showVeronese then
      dps.InitializeCameras(kWidthMainBox)
    else
      rist.InitializeCubeViewpointAndFocus;
    DrawScene;
  end;
  if clrs <> dps.numberOfColors then begin
    dps.numberOfColors:= clrs;
    if showVer then
      DrawScene;
  end;
  if showVer <> showVeronese then begin
    showVeronese:= showVer;
    DrawScene;
  end;
end;

procedure TImmersionR3R6.OnPaintWCursor(Sender: TObject);
begin
  DrawCursor;
end;


procedure TImmersionR3R6.OnManual(Sender: TObject);
begin
  Manual.Visible:= True;
end;

procedure TImmersionR3R6.OnChangeEditField(Sender: TObject);
{if the changed edit field contains something legible, redraw
 the scene}
  var
    tg, toChange: Integer;
begin
  if rist.busy then
    Exit;
  if not TEdit(Sender).modified then
    Exit;
  tg:= TEdit(Sender).tag;
  toChange:= 0;
  try
    ReadConstantField(compEdit[tg].text, xDom[tg]);
    if (min3D[tg] > xDom[tg]) then
      toChange:= 1
    else if (max3D[tg] < xDom[tg]) then
      toChange:= 2;
    inputStrs[16+tg]:= TEdit(sender).text;
    if toChange = 1 then begin
      min3D[tg]:= xDom[tg];
      inputStrs[9 + 2*tg]:= TEdit(sender).text;
      boundsEdit[tg,1].text:= TEdit(sender).text;
    end
    else if toChange = 2 then begin
      max3D[tg]:= xDom[tg];
      inputStrs[10 + 2*tg]:= TEdit(sender).text;
      boundsEdit[tg,2].text:= TEdit(sender).text;
    end;
    if toChange <> 0 then begin
      edges3D[tg]:= max3D[tg]-min3D[tg];
      halfEdges3D[tg]:= edges3D[tg]/2;
      center3D[tg]:= (max3D[tg]+min3D[tg])/2;
    end;
    if tg < 3 then
      DrawSpot(xDom)
    else begin
      wCrsPos:= round((xDom[3] - min3D[3])*kCrsWidth/edges3D[3]);
      DrawCursor;
    end;
    DrawScene;
    FillDomainWithFunction;
    domainBox.invalidate;
  except on exception do
    ;
  end
end;

procedure TImmersionR3R6.StopThreads;
{try to stop threads used in the computation of the Focal locus}
var
  i: Integer;
begin
  if thrNumber > 0 then
    for i:= 0 to numOfCPU-1 do
      if Assigned(thrs[i]) then
      thrs[i].Terminate;
  rist.busy:= False;
end;

procedure TImmersionR3R6.Wash;
begin
  beep(50,50);
  InfoBar.SimpleText:= eSingularPoint;
  domBmp.Canvas.FillRect(rctDomain);
  domainBox.canvas.draw(0,0,domBmp);
  thereIsSpot:= False;
  domBmp.Canvas.FillRect(rctDomain);
  domainBox.canvas.draw(0,0,domBmp);
  bmp.Canvas.FillRect(rctMain);
  MainBox.Canvas.draw(0,0,bmp);
  with SupplWindow  do begin
    if (tgOpt = Projective) then
      bm.canvas.Draw(0,0,bmStatic)
    else begin
      bm.canvas.Brush.Color:= clWhite;
      bm.canvas.FillRect(rct);
      bm.canvas.Pen.color:= $00CCCCCC;
      bm.canvas.Ellipse(rctReduced)
    end;
  end;
end;

end.


