unit FSupplWindow;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ExtCtrls, FDrawParamSurf, VectorsAndUtil, FImplicitCurveLine;

type
  TangentOption = (Sphere, Projective);
  DataOption = (Parabolic, Singular, Asymptotic);

  TSupplWindow = class(TForm)
    MainMenu1: TMainMenu;
    View: TMenuItem;
    Sphere1: TMenuItem;
    Projective1: TMenuItem;
    PaintBox1: TPaintBox;
    Data1: TMenuItem;
    ParabolicandsingularLoci1: TMenuItem;
    SingularLocus1: TMenuItem;
    AsymptoticLocus1: TMenuItem;
    procedure ComputeNewViewBasis(lastPt, curPt: TPoint);
    function DefinedRotVectors(p1, p2: TPoint; var e1,e2: Vector3): Boolean;
    procedure DrawContents;
    procedure DrawCurveProjective(crv: nodeArray; cl: TColor);
    procedure DrawCurves;
    procedure DrawCurveSphere(crv: nodeArray; cl: TColor);
    procedure DrawPointsProjective(na: nodeArray; cl: TColor);
    procedure DrawPointsSphere(na: nodeArray; cl: TColor);
    function FromParamsToBm(x: Vector2): TPoint;
    procedure OnMenuItems(Sender: TObject);
    procedure OnCreateForm(Sender: TObject);
    procedure OnData(Sender: TObject);
    procedure OnPaintBox(Sender: TObject);
    procedure OnDestroyForm(Sender: TObject);
    function StProjection(p: Vector2; var back: Boolean): PtSegment;
    procedure OnMouseMoveInForm(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OnMouseDownInForm(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure StoreSegmentAsympt(prevX, currentX: Vector2; key:Integer);
    procedure StoreSegmentParab(prevX, currentX: Vector2; key:Integer);
    procedure StoreSegmentSing1(prevX, currentX: Vector2; key:Integer);
    procedure StoreSegmentSing2(prevX, currentX: Vector2; key:Integer);
    procedure StoreSegmentSing3(prevX, currentX: Vector2; key:Integer);

  private
    c0, c1, c2, c3, cx, cy, sphRadius, ctr: Double;
    vert: pointDataGrid;
    min3DRed, max3DRed, ctr3DRed, vh, vv, vn: Vector3;
    lastPt: TPoint;
  public
    rct, rctReduced: TRect;
    drawWidth, offs: Integer;  {widht and height of the rectangle where the drawing is done}
    bm, bmStatic: TBitmap;
    parabLocus, asymptLocus{, asymptLocusSecDeg}: nodeArray;
    min2D, max2D: Vector2;
    pointList: nodeArray;
    tgOpt: TangentOption;
    dataOpt: DataOption;
    singF: array[1..3] of R2ToR1Function;
    singFgrad: array[1..3] of R2ToR2Function;
    singLocus: array[1..3] of nodeArray;
    storeSing: array[1..3] of StoreSegmentProcedure2D;
  end;

var
  SupplWindow: TSupplWindow;

implementation

{$R *.dfm}
uses
  Math,
  FRenderImplicitSurfaceWithThreads, ResourceStrings, FImmersionR3R6, FMapR3R6;

const
  cRevColor = $00BBBBBB;
  cRevRedColor = $00FFBBBB;
  cObvColor = $00FF00FF;
  cObvRedColor = $00BBBBFF;
  cCircColor= $000000FF;
  cDimmCircColor = $00AAAAFF;
  cl: array[1..3] of TColor =($00FF0000, $0000FF00, $000000FF);

procedure StorePoint(var na: nodeArray; v: Vector2);
begin
  inc(na.last);
  if high(na.xx) < na.last then
    SetLength(na.xx, na.last + 10);
  na.xx[na.last].x:= v;
end;

procedure TSupplWindow.ComputeNewViewBasis(lastPt, curPt: TPoint);
{Mouse was dragged from 'lastPt' to 'curPt'. Those points, by an
 inverse stereographic projection give two unit vectors that define
 a rotation around the axis orthogonal to both vectors passing by
 the origin, rotation that sends the first unit vector to the other.
 Here we use the decomposition of a rotation into two reflections,
 as was shown to me by Dick Palais, from Lehigh University. Of course,
 all this is done in the current basis of the plane where the projection
 is made, and we obtain the rotation by rotating the basis in the opposite
 sense}
var
  nn, twoVe2, vb, e2b: Double;
  e1, e2, b: Vector3;
  i: Integer;
begin
  if DefinedRotVectors(curPt, lastPt,  e1, e2) then begin
    for i := 1 to 3 do
      b[i] := e1[i] + e2[i];
      nn := EuclNorm(b);
      for i := 1 to 3 do
        b[i] := b[i] / nn;
      e2b := EuclInnerProd(e2,b);
      twoVe2 := 2 * EuclInnerProd(e2,vh);
      vb := EuclInnerProd(b,vh);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        vh[i] := vb * b[i] - twoVe2 * e2[i] + vh[i];
      twoVe2 := 2 * EuclInnerProd(e2,vv);
      vb := EuclInnerProd(b,vv);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        vv[i] := vb * b[i] - twoVe2 * e2[i] + vv[i];
      twoVe2 := 2 * EuclInnerProd(e2,vn);
      vb := EuclInnerProd(b,vn);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        vn[i] := vb * b[i] - twoVe2 * e2[i] + vn[i];
  end;
end;

function TSupplWindow.DefinedRotVectors(p1, p2: TPoint; var e1,e2: Vector3): Boolean;
{We consider that the screen in a plane by the origin of R^3 and that
 (vH, vV) is an orthonormal basis of that plane, so that the horizontal
 screen lines are parallel to vH and the vertical ones parallel to vV. Also
 vN is a unit vector orthogonal to that plane. For a point p in the screen
 we get the inverse stereographic projection on a sphere centered at the origin
 with radius 'sts.rEff'. Then we normalize the resulting vector. This we do
 with p1 and p2. The resulting points will define a rotation of R^3 that
 will be used to rotate the sphere.}
var
  t, x, y, d2, r, r2, sr, ph, pn, pv: double;
  i: Integer;
begin
  r:= sphRadius;
  r2:= sqr(r);
  x:= p1.x- ctr;  {components of p1 relative to the panel center}
  y:= p1.y- ctr;
  d2:= sqr(x)+ sqr(y);
  t:= 2*r2/(r2 + d2);

  {components of inverse stereographic projection of 'p1' in the basis
   (vHImg, vVImg, vNImg)}
  ph:= x*t;
  pv:= y*t;
  pn:= r*(1-t);

  for i:= 1 to 3 do
    e1[i]:= ph*vh[i]+ pv*vv[i]+ pn*vn[i];  {value of the inverse st. proj of p1}
  sr:= sqrt(sqr(e1[1])+ sqr(e1[2])+sqr(e1[3]));
  result:= (sr > 1.0e-5);
  if result then begin
    for i:= 1 to 3 do
      e1[i]:= e1[i]/sr;  {e1 becomes normalized, and we do the same with 'p2'}
    x:= p2.x- ctr;
    y:= p2.y- ctr;
    d2:= sqr(x)+ sqr(y);
    t:= 2*r2/(r2 + d2);
    ph:= x*t;
    pv:= y*t;
    pn:= r*(1-t);
    for i:= 1 to 3 do
      e2[i]:= ph*vh[i]+ pv*vv[i]+ pn*vn[i];
    sr:= sqrt(sqr(e2[1])+ sqr(e2[2])+sqr(e2[3]));
    result:= (sr > 1.0e-5);
    if result then begin
      for i:= 1 to 3 do
        e2[i]:= e2[i]/sr;
    end;
  end;
end;

procedure TSupplWindow.DrawContents;
var
  i: Integer;
begin
  if (tgOpt = Projective) then
    bm.canvas.Draw(0,0,bmStatic)
  else begin
    bm.canvas.Brush.Color:= clWhite;
    bm.canvas.FillRect(rct);
    bm.canvas.Pen.color:= $00CCCCCC;
    bm.canvas.Ellipse(rctReduced);
  end;

  if dataOpt = Parabolic then begin
    parabLocus.last:= -1;
    theImplicitCurveLine.ComputeImplicitCurve(MapR3R6.GaussCurvVeronese,
        nil, MapR3R6.CheckRangeStereogrCoord,
        StoreSegmentParab, min2D, max2D, 61, 600, 0.002);
    if tgOpt = Projective then
      DrawCurveProjective(parabLocus, clFuchsia)
    else
      DrawCurveSphere(parabLocus, clFuchsia);
  end
  else if dataOpt = Singular then
    for i:= 1 to 3 do begin
      singLocus[i].last:= -1;
      theImplicitCurveLine.ComputeImplicitCurve(singF[i], singFgrad[i],
        MapR3R6.CheckRangeStereogrCoord,
        storeSing[i], min2D, max2D, 61, 600, 0.001);
      if tgOpt = Projective then
        DrawCurveProjective(singLocus[i], cl[i])
      else if tgOpt = Sphere then
        DrawCurveSphere(singLocus[i], cl[i]);
    end
  else if dataOpt = Asymptotic then begin
    asymptLocus.last:= -1;
    pointList.last:= -1;
    theImplicitCurveLine.ComputeImplicitCurve(MapR3R6.AsymptoticFunctionStereogrCoord,
        MapR3R6.AsymptoticFunctionStereogrCoordGrad, MapR3R6.CheckRangeStereogrCoord,
        StoreSegmentAsympt, min2D, max2D, 41, 400, 0.002);
    {asymptLocusSecDeg.last:= -1;
    theImplicitCurveLine.ComputeImplicitCurve(MapR3R6.AsymptoticSecondDegreeInvar,
        nil, MapR3R6.CheckRangeStereogrCoord,
        StoreSegmentAsymptSecInv, min2D, max2D, 41, 400, 0.002);}
    if tgOpt = Projective then begin
      DrawCurveProjective(asymptLocus, clFuchsia);
      {DrawCurveProjective(asymptLocusSecDeg, clLime);}
      DrawPointsProjective(pointList, clNavy);
    end
    else begin
      DrawCurveSphere(asymptLocus, clFuchsia);
      {DrawCurveSphere(asymptLocusSecDeg, clLime);}
      DrawPointsSphere(pointList, clNavy);
    end;
  end;
  PaintBox1.canvas.Draw(0,0,bm);
end;

procedure TSupplWindow.DrawCurveProjective(crv: nodeArray; cl: TColor);
var
  i: Integer;
  pt: TPoint;
begin
  bm.canvas.Pen.color:= cl;
  if crv.last > 0 then begin
    for i:= 0 to crv.last do begin
      pt:= FromParamsToBm(crv.xx[i].x);
      if crv.xx[i].key = 0 then
        bm.Canvas.LineTo(pt.x, pt.y)
      else
        bm.Canvas.MoveTo(pt.x, pt.y)
    end;
  end;
end;

procedure TSupplWindow.DrawCurves;
var
  i: Integer;
begin
  if (tgOpt = Projective) then
    bm.canvas.Draw(0,0,bmStatic)
  else begin
    bm.canvas.Brush.Color:= clWhite;
    bm.canvas.FillRect(rct);
    bm.canvas.Pen.color:= $00CCCCCC;
    bm.canvas.Ellipse(rctReduced);
  end;

  if dataOpt = Parabolic then begin
    if tgOpt = Projective then
      DrawCurveProjective(parabLocus, clFuchsia)
    else
      DrawCurveSphere(parabLocus, clFuchsia);
  end
  else if dataOpt = Singular then
    for i:= 1 to 3 do begin
      if tgOpt = Projective then
        DrawCurveProjective(singLocus[i], cl[i])
      else if tgOpt = Sphere then
        DrawCurveSphere(singLocus[i], cl[i]);
    end
  else if dataOpt = Asymptotic then begin
      if tgOpt = Projective then begin
      DrawCurveProjective(asymptLocus, clFuchsia);
      {DrawCurveProjective(asymptLocusSecDeg, clLime);}
      DrawPointsProjective(pointList, clNavy);
    end
    else begin
      DrawCurveSphere(asymptLocus, clFuchsia);
      {DrawCurveSphere(asymptLocusSecDeg, clLime);}
      DrawPointsSphere(pointList, clNavy);
    end;
  end;
  PaintBox1.canvas.Draw(0,0,bm);
end;

procedure TSupplWindow.DrawCurveSphere(crv: nodeArray; cl: TColor);
var
  i: Integer;
  ptSgm, lastPtSgm: PtSegment;
  back, lastBack: Boolean;
  revColor: TColor;
begin
  if crv.last < 1 then
    Exit;
  revColor:= ((765+ cl shr 16) div 4) shl 16 + ((765+ (cl and $0000FF00) shr 8) div 4) shl 8+
              ((765+ (cl and $000000FF)) div 4);
  bm.canvas.Pen.color:= revColor;
  ptSgm:= StProjection(crv.xx[0].x, back);
  for i:= 1 to crv.last do begin
    lastPtSgm:= ptSgm;
    lastBack:= back;
    ptSgm:= StProjection(crv.xx[i].x, back);
    if (crv.xx[i].key = 0) then begin
      if ( back and lastBack ) then begin
        bm.canvas.moveto(lastPtSgm[1].x, lastPtSgm[1].y);
        bm.canvas.LineTo(PtSgm[1].x, PtSgm[1].y);
      end
      else begin
        bm.canvas.moveto(lastPtSgm[2].x, lastPtSgm[2].y);
        bm.canvas.LineTo(PtSgm[2].x, PtSgm[2].y);
      end
    end
  end;

  bm.canvas.Pen.color:= cl;
  ptSgm:= StProjection(crv.xx[0].x, back);
  for i:= 1 to crv.last do begin
    lastPtSgm:= ptSgm;
    lastBack:= back;
    ptSgm:= StProjection(crv.xx[i].x, back);
    if (crv.xx[i].key = 0) then begin
      if ( back and lastBack ) then begin
        bm.canvas.moveto(lastPtSgm[2].x, lastPtSgm[2].y);
        bm.canvas.LineTo(PtSgm[2].x, PtSgm[2].y);
      end
      else begin
        bm.canvas.moveto(lastPtSgm[1].x, lastPtSgm[1].y);
        bm.canvas.LineTo(PtSgm[1].x, PtSgm[1].y);
      end;
    end
  end;

  {bm.canvas.Brush.Color:= cDimmCircColor;
  if cvrsg.last >= 0 then begin
    for i:= 0 to cvrsg.last do begin
      ptSgm:= StProjection(cvrsg.xx[i].x, back);
      if back then
        bm.canvas.Ellipse(PtSgm[1].x+3, PtSgm[1].y+3, PtSgm[1].x-3, PtSgm[1].y-3)
      else
        bm.canvas.Ellipse(PtSgm[2].x+3, PtSgm[2].y+3, PtSgm[2].x-3, PtSgm[2].y-3);
    end;
  end;

  bm.canvas.Brush.Color:= cCircColor;
  if cvrsg.last >= 0 then begin
    for i:= 0 to cvrsg.last do begin
      ptSgm:= StProjection(cvrsg.xx[i].x, back);
      if back then
        bm.canvas.Ellipse(PtSgm[2].x+3, PtSgm[2].y+3, PtSgm[2].x-3, PtSgm[2].y-3)
      else
        bm.canvas.Ellipse(PtSgm[1].x+3, PtSgm[1].y+3, PtSgm[1].x-3, PtSgm[1].y-3)
    end;
  end;}
end;

procedure TSupplWindow.DrawPointsProjective(na: nodeArray; cl: TColor);
var
  i: Integer;
  pt: TPoint;
begin
  bm.canvas.Brush.color:= cl;
  bm.canvas.Pen.Color:= clWhite;
  if na.last >= 0 then begin
    for i:= 0 to na.last do begin
      pt:= FromParamsToBm(na.xx[i].x);
      bm.canvas.Ellipse(pt.x+3, pt.y+3, pt.x-3, pt.y-3)
    end;
  end;
end;

procedure TSupplWindow.DrawPointsSphere(na: nodeArray; cl: TColor);
var
  ptSgm: PtSegment;
  back: Boolean;
  revColor: TColor;
  i, j: Integer;
begin
  revColor:= ((765+ cl shr 16) div 4) shl 16 + ((765+ (cl and $0000FF00) shr 8) div 4) shl 8+
              ((765+ (cl and $000000FF)) div 4);
  bm.canvas.Pen.Color:= clWhite;
  if na.last >= 0 then begin
    for i:= 0 to na.last do begin
      ptSgm:= StProjection(na.xx[i].x, back);
      j:= 1 + ord(back);
      bm.canvas.Brush.color:= cl;
      bm.canvas.Ellipse(PtSgm[j].x+3, PtSgm[j].y+3, PtSgm[j].x-3, PtSgm[j].y-3);
      bm.canvas.Brush.color:= revColor;
      bm.canvas.Ellipse(PtSgm[3-j].x+3, PtSgm[3-j].y+3, PtSgm[3-j].x-3, PtSgm[3-j].y-3)
    end;
  end;
end;


function TSupplWindow.FromParamsToBm(x: Vector2): TPoint;
begin
  result.x:= round(offs + (x[1]+1)*drawWidth/2);
  result.y:= round(offs +  (1-x[2])*drawWidth/2);
end;

procedure TSupplWindow.OnCreateForm(Sender: TObject);
var
  i: Integer;
begin
  caption:= eSupplWindow;
  View.caption:= eView;
  Sphere1.caption:= eUnitSphereTg;
  Projective1.Caption:= eProjectiveTangent;
  SetRect(rct, 0, 0, PaintBox1.Width, PaintBox1.Width);
  drawWidth:= (12*PaintBox1.Width) div 13;
  offs:= (PaintBox1.Width - drawWidth) div 2;
  SetRect(rctReduced,offs, offs, offs+drawWidth, offs+drawWidth);

  tgOpt:= Sphere;
  ctr:= PaintBox1.Width div 2;
  sphRadius:= drawWidth div 2;
  bm:= TBitmap.Create;
  bm.Width:= PaintBox1.Width;
  bm.Height:= PaintBox1.Width;
  bm.PixelFormat:= pf32bit;

  bmStatic:= TBitmap.Create;
  bmStatic.Width:= PaintBox1.Width;
  bmStatic.Height:= PaintBox1.Width;
  bmStatic.PixelFormat:= pf32bit;
  with bmStatic.canvas do begin
    Pen.Color:= clWhite;
    Brush.Color:= $00DDFFDD;
    FillRect(rct);
    Brush.Color:= clWhite;
    Ellipse(rctReduced);
  end;

  vh[1]:= 1;
  vh[2]:= 0;
  vh[3]:= 0;
  vv[1]:= 0;
  vv[2]:= 1;
  vv[3]:= 0;
  vn[1]:= 0;
  vn[2]:= 0;
  vn[3]:= -1;

  min2D[1]:= -1;
  min2D[2]:= -1;
  max2D[1]:= 1;
  max2D[2]:= 1;

  for i:= 1 to 3 do begin
    singLocus[i].last:= -1;
    SetLength(singLocus[i].xx, 300);
  end;
  parabLocus.last:= -1;
  asymptLocus.last:= -1;
  {asymptLocusSecDeg.last:= -1;}
  SetLength(parabLocus.xx, 200);
  SetLength(asymptLocus.xx, 200);

  singF[1]:= MapR3R6.NormalCompVeronese1;
  singF[2]:= MapR3R6.NormalCompVeronese2;
  singF[3]:= MapR3R6.NormalCompVeronese3;
  singFgrad[1]:= MapR3R6.NormalCompVeronese1Grad;
  singFgrad[2]:= MapR3R6.NormalCompVeronese2Grad;
  singFgrad[3]:= MapR3R6.NormalCompVeronese3Grad;
  storeSing[1]:= StoreSegmentSing1;
  storeSing[2]:= StoreSegmentSing2;
  storeSing[3]:= StoreSegmentSing3;
  Show;
end;

procedure TSupplWindow.OnDestroyForm(Sender: TObject);
var
  i: Integer;
begin
  bm.Free;
  parabLocus.xx:= nil;
  asymptLocus.xx:= nil;
  for i:= 1 to 3 do
    singLocus[i].xx:= nil;
end;


procedure TSupplWindow.OnMenuItems(Sender: TObject);
var
  tgp: TangentOption;
begin
  if rist.busy then
    Exit;
  tgp:= TangentOption(TMenuItem(Sender).Tag);
  if tgp <> tgOpt then begin
    tgOpt:= tgp;
    DrawContents;
  end;
end;

procedure TSupplWindow.OnPaintBox(Sender: TObject);
begin
  PaintBox1.Canvas.Draw(0,0,bm);
end;

function TSupplWindow.StProjection(p: Vector2; var back: Boolean): PtSegment;
{Computes the inverse stereographic projection of 'p' to the unit sphere, and
 its opposite point, and the projects them to the current plane. Gives in back
 wheter result[1] is in front or in back of the viewer}
var
  a, aux: Double;
  v: Vector3;
begin
  aux:= sqr(p[1]) + sqr(p[2]);
  a:= sphRadius/(1 + aux);
  v[3]:= (1-aux)*a;
  a:= 2*a;
  v[1]:= a*p[1];
  v[2]:= a*p[2];
  aux:= EuclInnerProd(v,vH);
  result[1].x:= round(aux + ctr);
  result[2].x:= round(-aux + ctr);
  aux:= EuclInnerProd(v,vV);
  result[1].y:= round(ctr+ aux);
  result[2].y:= round(ctr- aux);
  back:= (EuclInnerProd(v,vN) < 0);
end;

procedure TSupplWindow.OnMouseMoveInForm(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  e1, e2: Vector3;
  pt: TPoint;
begin
  if (ssRight in Shift) and ((x <> lastPt.x) or (y <> lastPt.y)) and (tgOpt = Sphere) then begin
    pt:= Point(x,y);
    ComputeNewViewBasis(lastPt, pt);
    DrawCurves;
    lastPt:= pt;
  end;
end;

procedure TSupplWindow.OnMouseDownInForm(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if rist.busy then
    Exit;
  if (ssRight in Shift) then
    lastPt:= Point(x,y);
end;

procedure TSupplWindow.StoreSegmentAsympt(prevX, currentX: Vector2; key:Integer);
var
  X1, X2, X3, denInv, tr, aux: Double;
  v: Vector2;
  m: matrix33;
  a,b: Integer;
begin
  inc(asymptLocus.last);
  if asymptLocus.last > High(asymptLocus.xx) then
    SetLength(asymptLocus.xx, asymptLocus.last + 50);
  asymptLocus.xx[asymptLocus.last].key:= key;
  asymptLocus.xx[asymptLocus.last].x:= currentX;
  if (key = 0) and GetNearestZero22(MapR3R6.AsymptoticFunctionStereogrCoord,
          MapR3R6.AsymptoticSecondDegreeInvar,
          MapR3R6.AsymptoticFunctionStereogrCoordGrad,
          MapR3R6.AsymptoticSecondDegreeInvarGrad,
          prevX, currentX, 0.001, v)
        and (MapR3R6.TwistingMatrixRank(v, 1.0e-8) < 2) then
    StorePoint(pointList,v);
end;


procedure TSupplWindow.StoreSegmentParab(prevX, currentX: Vector2; key:Integer);
begin
  inc(parabLocus.last);
  if parabLocus.last > High(parabLocus.xx) then
    SetLength(parabLocus.xx, parabLocus.last + 50);
  parabLocus.xx[parabLocus.last].key:= key;
  parabLocus.xx[parabLocus.last].x:= currentX;
end;

procedure TSupplWindow.StoreSegmentSing1(prevX, currentX: Vector2; key:Integer);
begin
  inc(singLocus[1].last);
  if singLocus[1].last > High(singLocus[1].xx) then
    SetLength(singLocus[1].xx, singLocus[1].last + 50);
  singLocus[1].xx[singLocus[1].last].key:= key;
  singLocus[1].xx[singLocus[1].last].x:= currentX;
end;

procedure TSupplWindow.StoreSegmentSing2(prevX, currentX: Vector2; key:
    Integer);
begin
  inc(singLocus[2].last);
  if singLocus[2].last > High(singLocus[2].xx) then
    SetLength(singLocus[2].xx, singLocus[2].last + 50);
  singLocus[2].xx[singLocus[2].last].key:= key;
  singLocus[2].xx[singLocus[2].last].x:= currentX;
end;

procedure TSupplWindow.StoreSegmentSing3(prevX, currentX: Vector2; key:Integer);
begin
  inc(singLocus[3].last);
  if singLocus[3].last > High(singLocus[3].xx) then
    SetLength(singLocus[3].xx, singLocus[3].last + 50);
  singLocus[3].xx[singLocus[3].last].key:= key;
  singLocus[3].xx[singLocus[3].last].x:= currentX;
end;


procedure TSupplWindow.OnData(Sender: TObject);
var
  dtOpt: DataOption;
begin
  if rist.busy or not ImmersionR3R6.alreadyParsed then
    Exit;
  dtOpt:= DataOption(TMenuItem(sender).Tag);
  if dtOpt <> dataOpt then begin
    dataOpt:= dtOpt;
    DrawContents;
  end;
end;

end.
