
{-----------------------------------------------------------------------------

 Unit Name: VectorsAndUtil
 Project Name : ImmersionR3R6
 Author:    Angel Montesinos-Amilibia
 Date: 2009
Purpose: Collection of types and routines used along the project
-----------------------------------------------------------------------------}
{

ImmersionR3R6, workshop for the geometry of 3-submanifolds immersed in in R^6.

Copyright (C) 2009 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}


unit VectorsAndUtil;

interface
  uses
    Windows, Graphics;

const
  kIncr2Points = 1.4e-7;
  k2incr2Points= 2.8e-7;

  kIncr4Points= 0.0002;
  k2Incr4Points= 2*kIncr4Points;
  k12Incr4Points = 12*kIncr4Points;

  kTheError = 1511724757.0;  {a conventional number that I use. If a routine
      gives a result equal to 'kTheError', the caller knows that something was
      wrong}

  pi= 3.1415926535897932384;
  kInversePi:Double = 1/pi{3.1415926535897932384};
  kTwoPi = 2*pi;
  kFourPi = 4*pi;

  {kPPLine2D = 20;
  kMaxPPLine2D = 50;
  kMaxNumFlags = 4 * kMaxPPLine2D;}

{Geometry of the panels used in the form}

  kWidthMainBox= 496;
  kHalfWidth= kWidthMainBox div 2;
  kHeightMainBox= 496;
  kWidthDomBox= 224;
  kHeightDomBox = 224;
  kCrsWidth = 151;
  kSupplWWidth= 336;

{Other miscellaneous constants}
  kReturn = char(13);
  kEnter = char(3);
  kEscape = char(27);
  kTab = char($09);
  kPunto = '.';


type
  ByteArray= array of Byte;

  Vector2 = array[1..2] of Double;
  Vector3 = array[1..3] of Double;
  Vector6 = array[1..6] of Double;

  Vector2Single = array[1..2] of Single;
  Vector3Single = array[1..3] of Single;
  Vector6Single = array[1..6] of Single;

  Matrix0703 = array[0..7,0..3] of Double;

  Vector2Int = array[1..2] of Integer;
  Vector3Int = array[1..3] of Integer;
  Vector6Int = array[1..6] of Integer;

  PtSegment= array[1..2] of TPoint;
  Segment3 = array[1..2] of Vector3;
  Segment6 = array[1..2] of Vector6;

  matrix33 = array[1..3] of Vector3;
  matrix66 = array[1..6] of Vector6;
  matrix333 =  array[1..3, 1..3] of Vector3;

  R2ToR1Function = function(const pp: Vector2): Double of object;
  R2toR2Function = function(const pp: Vector2): Vector2 of object;
  MapR3R1= function (u: Vector3): Double of object;
  MapR3R3= function (u: Vector3): Vector3 of object;

  InRange3DFunction = function(var depth: Double; p: Vector3): Boolean of object;

  Paleta = ARRAY[0..255] OF TColor;
  PaletaPtr = ^Paleta;

  firstDeriv = record
    fa0: Double;
    dfa0: Double;
  end;

  secondDeriv = record  //used for the following: given a function in R^n, a
    fa0: Double;      //curve a:R --> R^n, f0 will be the value of
    dfa0 : Double;    //f(a(t)) at t=0, dfa0 will be f(a(t))'(0) (that is the
    ddfa0 : Double;   //value at t=0 of the derivative of f(a(t)); and ddfa0
  end;                  //will be f(a(t))''(0).

  DataBuffer = array of array of Single;
  wordMap = packed array of packed array of word;

  node= record
    key: Integer;
    x: Vector2;
  end;
  nodeArray = record
    last: Integer;
    xx: array of node;
  end;

  procedure ApplyGammaToMap;
  procedure ClearBitmap(var bm: TBitmap; h: Integer; c: TColor);
  function CrossAndNormalize(u,v: Vector3): Vector3;
  function CubicPolynomialRoot(a0,a1,a2,a3, t1, t2: Double; var t: Double): Boolean;  function CubicRoot(x: Double): Double;
  function Determinant3(m: matrix33): Double;
  function EuclCrossProd(u,v: Vector3): Vector3;
  function EuclDistance(x,y: Vector3): Double;
  function EuclEnergy(x: Vector3): Double; overload;
  function EuclEnergy(x: Vector6): Double; overload;
  function EuclInnerProd(u,v: Vector3): Double; overload;
  function EuclInnerProd(u,v: Vector6): Double; overload;
  function EuclNorm(v: Vector3): Double; overload;
  function EuclNorm(v: Vector6): Double; overload;
  function FormatFloatString(const x: Double; const totalStrLength: Integer): string;
  function GetNearestZero22(theF1, theF2: R2ToR1Function; theF1gr, theF2gr: R2ToR2Function;
              p1, p2: Vector2; rad: Double; var nz: Vector2): Boolean;
  function MatrixRank(var a: Matrix0703; m, n: integer; tol: Double): Integer;
  function MyArctan2(y,x: Double): Double;
  function Normalize(var v: Vector2): Boolean;
  procedure SetBufferSizes(w: Integer; fBufferAlso: Boolean);
  function SqrEuclDistance(x,y: Vector3): Double;

  var
    PaletaP: PaletaPtr;
    numOfCPU: Integer;
    bmp: TBitmap;
    map: wordMap;
    mapsRct: TRect;
    zBuffer, fBuffer: DataBuffer;

implementation

uses
  SysUtils, Math, FGammaPicker;

procedure ApplyGammaToMap;
{Apply the current color gamma to the surface bitmap}
  var
    i, j: Integer;
    line: pTColorArray;
    c: TColor;
begin
  {'c' is background color}
  c:= gammas[0,0];
  bmp.Canvas.Brush.Color:= (c and $00FF0000) shr 16 + (c and $0000FF00)+ (c and $000000FF) shl 16;
  for j:= 0 to mapsRct.bottom-1 do begin
    line:= bmp.ScanLine[j];
    for i:= 0 to mapsRct.right-1 do
      line^[i]:= gammas[ (map[i,j] and $8000) shr 15, map[i,j] and $7FFF];
  end;
end;

procedure ClearBitmap(var bm: TBitmap; h: Integer; c: TColor);
{Fill 'bm' with the color 'c', 'h' is the number of pixels that will
 be filled in each row of 'bm'}
var
  line: pTColorArray;
  i,j: Integer;
begin
  for i:=0 to h-1 do begin
    line:= bm.ScanLine[i];
    for j:= 0 to h-1 do
      line^[j]:= c;
  end;
end;

function CrossAndNormalize(u,v: Vector3): Vector3;
{makes the cross product of 'u' and 'v' and normalizes the
result. If it fails, returns result[1]:= 3}
var
  aux: Double;
  i: Integer;
begin
  result[1]:= u[2]*v[3] - u[3]*v[2];
  result[2]:= u[3]*v[1] - u[1]*v[3];
  result[3]:= u[1]*v[2] - u[2]*v[1];
  aux:= EuclNorm(result);
  if aux > 1.0e-20 then begin
    aux:= 1/aux;
    for i:= 1 to 3 do
      result[i]:= aux*result[i];
  end
  else
    result[1]:= 3;
end;

function CubicPolynomialRoot(a0,a1,a2,a3, t1, t2: Double; var t: Double): Boolean;
{Gives in 't' the smallest root of the polynomial a0 + a1*t + a2*t^2 + a3*t^3 in the
interval [t1,t2]. Result is True if it founds one}
var
   b, c, d, p, q, p3, u, v, discr, phi: Double;
   roots: Vector3;
   i: Integer;
begin
  result:= False;
  if t1 > t2 then
    Exit;
  t:= t2 + 1;
  if a3 <> 0 then begin
    b:= a2/a3;
    c:= a1/a3;
    d:= a0/a3;
    p:= (3*c - sqr(b))/9;
    q:= (2*b*sqr(b)-9*b*c + 27*d)/54;
    p3:= p*sqr(p);
    discr:= p3 + sqr(q);
    if discr < 0 then begin
      phi:= arccos(-q/sqrt(abs(p3)));
      p:= 2*sqrt(abs(p));
      roots[1]:= p*cos(phi/3)-b/3;
      roots[2]:= -p*cos((phi+ pi)/3)-b/3;
      roots[3]:= -p*cos((phi-pi)/3)-b/3;
      for i:= 1 to 3 do
        if (roots[i] >= t1) and (roots[i] <= t2) and (roots[i] < t) then
          t:= roots[i];
      result:= (t <= t2);
    end
    else begin
      discr:= sqrt(discr);
      u:= cubicRoot(-q + discr);
      v:= cubicRoot(-q - discr);
      t:= u + v - b/3;
      result:= (t >= t1) and (t <= t2);
    end;
  end
  else begin
    if (a2 <> 0) then begin
      discr:= sqr(a1)-4*a0*a2;
      result:= (discr >= 0);
      if result then begin
        discr:= sqrt(discr);
        roots[1]:= (-a1+discr)/(2*a2);
        roots[2]:= (-a1-discr)/(2*a2);
        for i:= 1 to 2 do
          if (roots[i] >= t1) and (roots[i] <= t2) and (roots[i] < t) then
            t:= roots[i];
        result:= (t <= t2);
      end;
    end
    else begin
      result:= (a1 <> 0);
      if result then begin
        t:= -a0/a1;
        result:= (t >= t1) and (t <= t2);
      end;
    end;
  end;
end;



function CubicRoot(x: Double): Double;
begin
  if abs(x) < 1.0e-60 then
    result:= 0
  else begin
    result:= exp(ln(abs(x))/3);
    if x < 0 then
      result:= - result;
  end;
end;

function Determinant3(m: matrix33): Double;
begin
  result:= m[1,1]*(m[2,2]*m[3,3]-m[2,3]*m[3,2]) +
           m[2,1]*(m[3,2]*m[1,3]-m[3,3]*m[1,2]) +
           m[3,1]*(m[1,2]*m[2,3]-m[1,3]*m[2,2]);
end;

function EuclCrossProd(u,v: Vector3): Vector3;
begin
  result[1]:= u[2]*v[3] - u[3]*v[2];
  result[2]:= u[3]*v[1] - u[1]*v[3];
  result[3]:= u[1]*v[2] - u[2]*v[1];
end;

function EuclDistante(x,y: Vector6): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3])+sqr(x[4]-y[4])+
                sqr(x[5]-y[5])+ sqr(x[6]-y[6]));
end;

function EuclDistance(x,y: Vector3): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3]));
end;



function EuclEnergy(x: Vector3): Double; overload;
begin
  result:= sqr(x[1])+ sqr(x[2])+sqr(x[3]);
end;

function EuclEnergy(x: Vector6): Double;
begin
  result:= sqr(x[1])+ sqr(x[2])+sqr(x[3])+sqr(x[4])+ sqr(x[5]) + sqr(x[6]);
end;


function EuclInnerProd(u,v: Vector3): Double; overload;
begin
  result:= u[1]*v[1] + u[2]*v[2] + u[3]*v[3];
end;


function EuclInnerProd(u,v: Vector6): Double; overload;
begin
  result:= u[1]*v[1] + u[2]*v[2] + u[3]*v[3] + u[4]*v[4] + u[5]*v[5] + u[6]*v[6];
end;

function EuclNorm(v: Vector3): Double; overload;
begin
  result:= sqrt(sqr(v[1]) + sqr(v[2]) + sqr(v[3]));
end;

function EuclNorm(v: Vector6): Double; overload;
begin
  result:= sqrt(sqr(v[1]) + sqr(v[2]) + sqr(v[3]) + sqr(v[4])+ sqr(v[5]) + sqr(v[6]));
end;

function FormatFloatString(const x: Double; const totalStrLength: Integer): string;
  var
    sigDig: Integer;
begin
  result:= FloatToStrF(x, ffFixed, totalStrLength+2,15);
  sigDig:= Length(result);
  if sigDig > totalStrLength  then
    sigDig:= totalStrLength;
  result:= Copy(result,1,sigDig);
end;

function GetNearestZero22(theF1, theF2: R2ToR1Function; theF1gr, theF2gr: R2ToR2Function;
 p1, p2: Vector2; rad: Double; var nz: Vector2): Boolean;
{Initially, 'p1' and 'p2' are close points in R^2 on which 'theF1' vanishes,
 and such that on them 'theF2' has values of opposite signs. We look for a nearby
 point on which both functions vanish and return it in 'nz'. Iff we find it,
 result is True; otherwise it is False. 'theF1gr' and 'theF2gr' are the respective
 gradients of 'theF1' and 'theF2'}
var
  maxShift, f1, f2, f, absF, absFprev, shift, aux: Double;
  pos1, pos2, gr1, gr2, h, v: Vector2;
  i: Integer;
begin
  maxShift:= 2*hypot(p1[1]-p2[1], p1[2]-p2[2]);
  if maxShift > 1.0e-3*rad then begin
    pos1:= p1;
    pos2:= p2;
    f1:= theF2(pos1);
    f2:= theF2(pos2);
    if f1*f2 < 0 then begin
      for i:= 1 to 2 do
        v[i]:= (pos1[i]+ pos2[i])/2;
      f:= theF2(v);
      while abs(f) > 1.0e-12 do begin
        if (f1*f) >= 0 then begin
          pos1:= v;
          f1:= f;
        end
        else begin
          pos2:= v;
          f2:= f;
        end;
        for i:= 1 to 2 do
          v[i]:= (pos1[i]+ pos2[i])/2;
        f:= theF2(v);
      end;
      f2:= f;
      f1:= theF1(v);
      absF:= abs(f1) + abs(f2);
      absFPrev:= 2*absF;
      shift:= 0;
      while (absFPrev > absF) and (shift < maxShift) do begin
        nz:= v;
        absFPrev:= absF;
        gr1:= theF1gr(v);
        gr2:= theF2gr(v);
        aux:= gr1[1]*gr2[2] - gr1[2]*gr2[1];
        if abs(aux) > 1.0e-16 then begin
          h[1]:= (f2*gr1[2] - f1*gr2[2])/aux;
          h[2]:= (f1*gr2[1] - f2*gr1[1])/aux;
          v[1]:= v[1] + h[1];
          v[2]:= v[2] + h[2];
          shift:= abs(h[1]) + abs(h[2]);
        end
        else
          break;
        f1:= theF1(v);
        f2:= theF2(v);
        absF:= abs(f1)+ abs(f2);
      end;
      absF:= min(absF, absFPrev);
    end
    else begin
      for i:= 1 to 2 do
        nz[i]:= (pos1[i]+ pos2[i])/2;
      absF:= abs(theF1(nz)) + abs(theF2(nz));
    end;
  end
  else
    absF:= abs(theF1(p1)) + abs(theF2(p1));
  result:= (absF < 1.0e-8);
end;

function MatrixRank(var a: Matrix0703; m, n: integer; tol: Double): Integer;
{Computes the rank of the matrix a[i,j], where 1<= i <= m, and 1<= j <= n are
 the effective values.
 The dependencies are signalled when the absolute value of certain algebraic
 expressions are less than 'tol'}
const
  hugeValue = 1.0e100;
   var
    md, i, j, k, h, i0, j0 : Integer;
    b, minVal: Double;
    c, v: array of Double;
begin
  result:= 0;
  md:= m;
  if n < m then
    md:= n;
  SetLength(c,m+1);
  SetLength(v,n+1);
  while result < md do begin
    minVal:= hugeValue;
    i0 := 0;
    j0 := 0;
    for i:= result+1 to m do
      for j:= result+1 to n do
        if (abs(a[i,j]) > tol) and ( abs(a[i,j]) < minVal ) then begin
          minVal:= abs(a[i,j]);
          i0 := i;
          j0 := j;
        end;
    if (i0 = 0) and (j0 = 0) then begin
      c := nil;
      Exit;
    end;
    if ( result+1 < md ) then begin
      for i := result+1 to m do begin
        c[i] := a[i, result+1];
        a[i, result+1] := a[i,j0];
        a[i,j0] := c[i];
      end;
      while (i0 > 0) do begin
        for h:= 1 to n do
          v[h] := a[result+1,h];
        a[result+1] := a[i0];
        for h:= 1 to n do
          a[i0,h]:= v[h];
        i0 := 0;
        minVal := hugeValue;
        for i := result+2 to m do begin
          if ( abs(a[i,result+1]) > tol ) then begin
            b:= a[i,result+1]/a[result+1, result+1];
            for k:= result+1 to n do
              a[i,k] := a[i,k] - b*a[result+1,k];
          end;
          if ( abs(a[i,result+1]) > tol ) and ( abs(a[i,result+1]) < minVal ) then begin
            minVal := abs(a[i,result+1]);
            i0 := i;
          end;
        end;
      end;
    end;
    inc(result);
  end;
  c:= nil;
  v:= nil;
end;

function MyArctan2(y,x: Double): Double;
begin
  if abs(y) < 1.0e16 * abs(x) then
    result:= arctan2(y,x)
  else if (y > 0) then
      result:= pi/2
  else if ( y < 0 ) then
      result:= -pi/2
  else
      result:= 0;
end;

function Normalize(var v: Vector2): Boolean;
var
  a: Double;
begin
  a:= sqr(v[1]) + sqr(v[2]);
  result:=  a > 1.0e-30;
  if result then begin
    a:= 1/ sqrt(a);
    v[1]:= v[1]*a;
    v[2]:= v[2]*a;
  end;
end;

procedure SetBufferSizes(w: Integer; fBufferAlso: Boolean);
var
  i: Integer;
begin
  SetLength(zBuffer,w);
  for i:= 0 to w-1 do
    SetLength(zBuffer[i], w);

  SetLength(fBuffer,w);
  for i:= 0 to w-1 do
    SetLength(fBuffer[i], w);

  SetLength(map, w);
  for i:= 0 to w-1 do
    SetLength(map[i], w);
end;

function SqrEuclDistance(x,y: Vector3): Double;
begin
  result:= sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3]);
end;

procedure Initialize;
var
  i: Integer;
begin
  bmp:= TBitmap.Create;
  bmp.Width:= kWidthMainBox;
  bmp.Height:= kWidthMainBox;
  bmp.PixelFormat := pf32bit;
  bmp.Canvas.Brush.Color := clWhite;
  bmp.Canvas.Brush.Style := bsSolid;
  SetLength(zBuffer,kWidthMainBox);
  SetLength(fBuffer,kWidthMainBox);
  SetLength(map,kWidthMainBox);
  SetRect(mapsRct,0,0,kWidthMainBox,kWidthMainBox);
  for i:= 0 to kWidthMainBox-1 do begin
    SetLength(zBuffer[i],kWidthMainBox);
    SetLength(fBuffer[i],kWidthMainBox);
    SetLength(map[i],kWidthMainBox);
  end;
end;


procedure Finalize;
var
  i: Integer;
begin
  bmp.Free;
  if high(zBuffer)>0 then begin
    for i:= 0 to High(zBuffer) do
      zBuffer[i]:= nil;
    zBuffer:= nil;
  end;
  if high(zBuffer)>0 then begin
    for i:= 0 to High(zBuffer) do
      zBuffer[i]:= nil;
    zBuffer:= nil;
  end;
  if high(map)>=0 then begin
    for i:= 0 to High(map) do
      map[i]:= nil;
    map:= nil;
  end;
end;


initialization
  initialize;

finalization
  finalize;


END.
 
